# Runs steps for validation

import lsmtool


s = lsmtool.load('tests/no_patches.sky')

print('Select individual sources with Stokes I fluxes above 1 Jy')
s.select('I > 1.0 Jy')

print('Transfer patches from patches.sky')
s.transfer('tests/patches.sky')

print('Remove patches with total fluxes below 2 Jy')
s.remove('I < 2.0 Jy', aggregate='sum')

print('Ungroup the skymodel')
s.ungroup()

print('Concatenate with concat.sky')
s.concatenate('tests/concat.sky', matchBy = 'position', radius = '30 arcsec', keep = 'from2')

print('Compare to concat.sky')
c = lsmtool.load('tests/concat.sky')
c.ungroup()
c.select('I > 5.0 Jy')
s.ungroup()
s.compare(c, outDir='tests/')

print('Add a source')
s.add({'Name': 'src1', 'Type': 'POINT', 'Ra': 277.4232, 'Dec': 48.3689, 'I': 0.69})

print('Group using tessellation to a target flux of 50 Jy')
s.group('tessellate', targetFlux = '50.0 Jy')

print('Move patch Patch_1 to 16:04:16.2288, 58.03.06.912')
s.move('Patch_1', position =  ['16:04:16.2288', '58.03.06.912'])

print('Merge patches Patch_0 and Patch_2')
s.merge(['Patch_0', 'Patch_2'], name = 'merged_patch')

print('Reset patch positions to midpoint of patch and write final model to file')
s.setPatchPositions(method='mid')
s.write('tests/final.sky', clobber=True)

print('Plot the sky model')
s.plot('tests/plot.pdf', labelBy='patch')



from os.path import dirname, basename, isfile
from glob import glob

modules = glob(dirname(__file__)+"/*.py")
__all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py') and not basename(f).startswith('_')]
del modules

from . import *




# -*- coding: utf-8 -*-
#
# This module stores the version and changelog

# Version number
__version__ = '1.4.11'

# Change log
def changelog():
    log = """
    LSMTool Changelog.
    -----------------------------------------------

    2023/02/10 - Version 1.4.11

        Replace all deprecated `numpy` types with their plain counterpart

    2023/01/09 - Version 1.4.10

        Fix issues with concatenating two sky models

        Fix beam attenuation when using EveryBeam

        Fix handling of LogarithmicSI and OrientationIsAbsolute columns

        Fix skymodel compare operation

        Replace deprecated `numpy.float` with plain `float`

    2022/08/29 - Version 1.4.9

        Documentation moved to https://lsmtool.readthedocs.io.

        Repository moved to https://git.astron.nl/RD/LSMTool.git,
        https://github.com/darafferty/LSMTool.git is now a mirror.

        Added a new script correct_gaussian_orientation.py, which computes
        the absolute Gaussian position angle for a skymodel file.

        Fixed numpy deprecation warnings.

        A few minor bug fixes.

    2022/06/09 - Version 1.4.8

        Fix bug in concatenation of sky models when number of spectral index
        terms differs

    2022/02/23 - Version 1.4.7

        Fix reading of patchless tables for astropy versions < 4.1

    2022/02/08 - Version 1.4.6

        Fix table column converters for astropy versions >= 4.1

    2022/02/04 - Version 1.4.5

        Fix regression in handling of blank values introduced in v1.4.4

    2022/01/27 - Version 1.4.4

        Fix improper handling of blank values

        Add support for EveryBeam

    2021/04/07 - Version 1.4.3

        Publish on PyPI

        Add faster version of meanshift algorithm

        Fix to incorrect filtering with mask images

        Fix to Astropy registry check

    2019/10/01 - Version 1.4.2

        Fix to incorrect header on write

        Add meanshift grouping algorithm

    2019/04/12 - Version 1.4.1

        Fix installation on Python 3 systems

        Update GSM url

    2019/03/21 - Version 1.4.0

        Allow 2D images as masks in group operation

        Add TGSS and GSM VO queries

        Add option to apply beam attenuation during write

    2018/05/04 - Version 1.3.1

        Improve packaging support

    2018/05/03 - Version 1.3.0

        Add option to tessellate using patches

    2017/06/15 - Version 1.2.0

    2017/05/09

        Add a "pad_index" option to lsm.group() to allow the patch index to be
        padded with leading zeros

    2017/05/05

        Improve the "tessellate" grouping algorithm when negative fluxes are
        present

    2017/04/21

        Add support for LogarithmicSI column

        Add a "voronoi" option to lsm.group() to allow a previously grouped sky
        model to be regrouped using Voronoi tessellation

    2016/01/28 - Version 1.1.0

        Add support for FACTOR-formatted output

    2014/06/25 - Version 1.0.0 (initial release)
    """

    print(log)



# -*- coding: utf-8 -*-
#
# Defines cluster functions used by the group operation
#
# Copyright (C) 2013 - Reinout van Weeren
# Copyright (C) 2013 - Francesco de Gasperin
# Modified by David Rafferty as required for integration into LSMTool
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import sys
import numpy as np
import logging
import itertools

class Patch():
    def __init__(self, name, ra, dec, flux):
        self.name = name
        self.ra = ra
        self.dec = dec
        self.flux = flux

class Cluster():
    def __init__(self, name, patch):
        self.name = name
        self.init_patch = patch
        self.patches = []
        self.update_cluster_coords()

    def add_patch(self, patch):
        """
        Add a patch to this cluster
        """
        self.patches.append(patch)

    def total_cluster_flux(self):
        """
        Return total cluster flux
        """
        cluster_flux = self.init_patch.flux
        for patch in self.patches:
            cluster_flux += patch.flux

        return cluster_flux

    def update_cluster_coords(self):
        """
        update the self.centroid_ra, self.centroid_dec, self.mid_ra, self.mid_dec
        """
        self.centroid_ra = self.init_patch.ra*self.init_patch.flux
        self.centroid_dec = self.init_patch.dec*self.init_patch.flux
        min_ra = np.inf
        max_ra = -np.inf
        min_dec = np.inf
        max_dec = -np.inf

        for patch in self.patches:
            self.centroid_ra += patch.ra*patch.flux
            self.centroid_dec += patch.dec*patch.flux
            if patch.ra < min_ra: min_ra = patch.ra
            if patch.ra > max_ra: max_ra = patch.ra
            if patch.dec < min_dec: min_dec = patch.dec
            if patch.dec > max_dec: max_dec = patch.dec

        self.centroid_ra /= self.total_cluster_flux()
        self.centroid_dec /= self.total_cluster_flux()
        self.mid_ra = min_ra + (max_ra - min_ra)/2.
        self.mid_dec = min_dec + (max_dec - min_dec)/2.


def ratohms_string(ra):
    rah, ram, ras = ratohms(ra)
    return str(rah) + ':' + str(ram) + ':' +str(round(ras,2))


def dectodms_string(dec):
    decd, decm, decs = dectodms(dec)
    return str(decd) + '.' + str(decm) + '.' +str(round(decs,2))


def compute_patch_center(LSM, applyBeam=False):
    """
    Return the patches names, central (weighted) RA and DEC and total flux
    """
    data = LSM.table
    patch_names = np.unique(data['Name'])
    fluxes = LSM.getColValues('I', applyBeam=applyBeam)
    patches = []

    for patch_name in patch_names:
        comp_ids = np.where(data['Name'] == patch_name)[0]

        patch_ra    = 0.
        patch_dec   = 0.
        weights_ra  = 0.
        weights_dec = 0.
        patch_flux  = 0.
        for comp_id in comp_ids:

            comp_ra   = data['Ra'][comp_id]
            comp_dec  = data['Dec'][comp_id]
            comp_flux = fluxes[comp_id]

            # calculate the average weighted patch center, and patch flux
            patch_flux  += comp_flux
            patch_ra    += comp_ra * comp_flux
            patch_dec   += comp_dec * comp_flux
            weights_ra  += comp_flux
            weights_dec += comp_flux

        patches.append(Patch(patch_name, patch_ra/weights_ra, patch_dec/weights_dec, patch_flux))

    return patches


def angsep2(ra1, dec1, ra2, dec2):
    """Returns angular separation between two coordinates (all in degrees)"""
    from astropy.coordinates import FK5
    import astropy.units as u

    coord1 = FK5(ra1, dec1, unit=(u.degree, u.degree))
    coord2 = FK5(ra2, dec2, unit=(u.degree, u.degree))

    return coord1.separation(coord2)


def create_clusters(LSM, patches_orig, Q, applyBeam=False, root='Patch', pad_index=False):
    """
    Clusterize all the patches of the skymodel iteratively around the brightest patches
    """
    from astropy.coordinates import SkyCoord
    from astropy import units as u
    from distutils.version import LooseVersion
    import scipy
    log = logging.getLogger('LSMTool.Cluster')
    if LooseVersion(scipy.__version__) < LooseVersion('0.11.0'):
        log.debug('The installed version of SciPy contains a bug that affects catalog matching. '
            'Falling back on (slower) matching script.')
        from ._matching import match_coordinates_sky
    else:
        from astropy.coordinates.matching import match_coordinates_sky

    # sort the patches by brightest first
    idx = np.argsort([patch.flux for patch in patches_orig])[::-1] # -1 to reverse sort
    patches = list(np.array(patches_orig)[idx])

    # initialize clusters with the brightest patches
    clusters = []
    for i, patch in enumerate(patches[0:Q]):
        if pad_index:
            clusters.append(Cluster(root+'_'+str(i).zfill(int(np.ceil(np.log10(Q)))), patch))
        else:
            clusters.append(Cluster(root+'_'+str(i), patch))

    # Iterate until no changes in which patch belongs to which cluster
    count = 1
    patches_seq_old = []
    patchRAs = []
    patchDecs = []
    for patch in patches:
        patchRAs.append(patch.ra)
        patchDecs.append(patch.dec)

    while True:
        clusterRAs = []
        clusterDecs = []
        if LSM.hasPatches:
            clusterRA, clusterDec = LSM.getPatchPositions(method='wmean',
                asArray=True, applyBeam=applyBeam, perPatchProjection=False)
            clusterNames = LSM.getPatchNames()
            patches_orig = LSM.getColValues('Name')
        else:
            clusterRA = [cluster.centroid_ra for cluster in clusters]
            clusterDec = [cluster.centroid_dec for cluster in clusters]
            clusterNames = [cluster.name for cluster in clusters]
        for cluster in clusters:
            # reset patches
            if type(clusterNames) is not list:
                clusterNames = clusterNames.tolist()
            cindx = clusterNames.index(cluster.name)
            cluster.patches = []
            clusterRAs.append(clusterRA[cindx])
            clusterDecs.append(clusterDec[cindx])

        catalog1 = SkyCoord(clusterRAs, clusterDecs,
            unit=(u.degree, u.degree), frame='fk5')
        catalog2 = SkyCoord(patchRAs, patchDecs,
            unit=(u.degree, u.degree), frame='fk5')
        matchIdx, d2d, d3d = match_coordinates_sky(catalog2, catalog1)

        for i, patch in zip(matchIdx, patches):
            cluster = clusters[i]
            cluster.add_patch(patch)

        patches_seq = []
        for cluster in clusters:
            patches_seq.extend(cluster.patches)

        count += 1
        if patches_seq == patches_seq_old:
            break
        patches_seq_old = patches_seq

        # Make output patch column
        patchNames = [''] * len(patches)
        patchNames_orig = LSM.getColValues('Name').tolist()
        for c in clusters:
            for p in c.patches:
                patchNames[patchNames_orig.index(p.name)] = c.name

        LSM.setColValues('Patch', patchNames, index=2)
        LSM._updateGroups()
    return np.array(patchNames)




# -*- coding: utf-8 -*-
#
# This module defines functions used for more than one operation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging
from astropy.coordinates import Angle
from math import floor, ceil


def normalize_ra(ang):
    """
    Normalize RA to be in the range [0, 360).

    Based on https://github.com/phn/angles/blob/master/angles.py

    Parameters
    ----------
    ang : float or astropy.coordinates.Angle
        The RA in degrees to be normalized.

    Returns
    -------
    res : float
        RA in degrees in the range [0, 360).
    """
    lower = 0.0
    upper = 360.0
    if type(ang) is Angle:
        num = ang.value
    else:
        num = ang
    res = num
    if num > upper or num == lower:
        num = lower + abs(num + upper) % (abs(lower) + abs(upper))
    if num < lower or num == upper:
        num = upper - abs(num - lower) % (abs(lower) + abs(upper))
    res = lower if num == upper else num

    return res


def normalize_dec(ang):
    """
    Normalize Dec to be in the range [-90, 90].

    Based on https://github.com/phn/angles/blob/master/angles.py

    Parameters
    ----------
    ang : float or astropy.coordinates.Angle
        The Dec in degrees to be normalized.

    Returns
    -------
    res : float
        Dec in degrees in the range [-90, 90].
    """
    lower = -90.0
    upper = 90.0
    if type(ang) is Angle:
        num = ang.value
    else:
        num = ang
    res = num
    total_length = abs(lower) + abs(upper)
    if num < -total_length:
        num += ceil(num / (-2 * total_length)) * 2 * total_length
    if num > total_length:
        num -= floor(num / (2 * total_length)) * 2 * total_length
    if num > upper:
        num = total_length - num
    if num < lower:
        num = -total_length - num
    res = num

    return res


def radec_to_xyz(ra, dec, time):
    """
    Convert RA and Dec coordinates to ITRS coordinates for LOFAR observations.

    Note: the location adopted for LOFAR is that of the core, so the returned
    ITRS coordinates will be less accurate for stations outside of the core.

    Parameters
    ----------
    ra : astropy.coordinates.Angle
        Right ascension
    dec: astropy.coordinates.Angle
        Declination
    time: float
        MJD time in seconds

    Returns
    -------
    pointing_xyz: array
        NumPy array containing the ITRS X, Y and Z coordinates
    """
    from astropy.coordinates import SkyCoord, ITRS, EarthLocation, AltAz
    from astropy.time import Time
    import astropy.units as u
    import numpy as np

    obstime = Time(time/3600/24, scale='utc', format='mjd')

    # Set the location to the LOFAR core
    loc_LOFAR = EarthLocation(lon=0.11990128407256424, lat=0.9203091252660295, height=6364618.852935438*u.m)

    dir_pointing = SkyCoord(ra, dec)
    dir_pointing_altaz = dir_pointing.transform_to(AltAz(obstime=obstime, location=loc_LOFAR))
    dir_pointing_xyz = dir_pointing_altaz.transform_to(ITRS)

    pointing_xyz = np.asarray([dir_pointing_xyz.x, dir_pointing_xyz.y, dir_pointing_xyz.z])
    return pointing_xyz


def apply_beam_star(inputs):
    """
    Simple multiprocessing helper function for apply_beam()
    """
    return apply_beam(*inputs)


def apply_beam(RA, Dec, spectralIndex, flux, referenceFrequency, beamMS, time, ant1,
               numchannels, startfreq, channelwidth, invert):
    """
    Worker function for attenuate() that applies the beam to a single flux

    Parameters
    ----------
    RA : float
        RA value in degrees
    Dec : float
        Dec value in degrees
    spectralIndex : float
        Spectral index to adjust
    flux : list
        Flux to attenuate
    referenceFrequency : float
        Reference frequency of polynomial fit
    beamMS : str
        Measurement set for which the beam model is made
    time : float
        Time to calculate beam (in MJD seconds)
    ant1 : int
        Station index
    numchannels : int
        Number of channels
    startfreq : float
        Start frequency in Hz
    channelwidth : float
        Channel with in Hz
    invert : bool
        If True, invert the beam (i.e. to un-attenuate the flux)

    Returns
    -------
    attFlux : float
        Attenuated flux
    adjSpectralIndex : float
        Adjusted spectral index. Returned only if spectralIndex is not None

    """
    import numpy as np
    import everybeam as eb
    import casacore.tables as pt
    from astropy.coordinates import Angle
    import astropy.units as u

    # Use ant1, times, and n channel to compute the beam, where n is determined by
    # the order of the spectral index polynomial
    sr = eb.load_telescope(beamMS)
    source_ra = Angle(RA, unit=u.deg)
    source_dec = Angle(Dec, unit=u.deg)
    source_xyz = radec_to_xyz(source_ra, source_dec, time)
    obs = pt.table(beamMS+'::FIELD', ack=False)
    pointing_ra = Angle(float(obs.col('REFERENCE_DIR')[0][0][0]), unit=u.rad)
    pointing_dec = Angle(float(obs.col('REFERENCE_DIR')[0][0][1]), unit=u.rad)
    obs.close()
    pointing_xyz = radec_to_xyz(pointing_ra, pointing_dec, time)
    freqs = np.arange(startfreq, startfreq+numchannels*channelwidth, channelwidth)

    # Correct the source spectrum if needed
    if spectralIndex is not None:
        nspec = len(spectralIndex)
        s = max(1, int(numchannels/nspec))
        ch_indices = list(range(0, numchannels, s))
        ch_indices.append(numchannels)
    else:
        ch_indices = [int(numchannels/2)]
    freqs_new = []
    fluxes_new = []
    for ind in ch_indices:
        # Evaluate beam and take XX only (XX and YY should be equal) and square
        freq = freqs[ind]
        beam = abs(sr.array_factor(time, ant1, freq, source_xyz, pointing_xyz))
        if invert:
            beam = 1 / beam
        beam = beam[0][0]**2
        if spectralIndex is not None:
            nu = (startfreq + ind*channelwidth) / referenceFrequency - 1
            freqs_new.append(nu)
            fluxes_new.append(polynomial(flux, spectralIndex, nu) * beam)
        else:
            fluxes_new.append(flux * beam)
    if spectralIndex is not None:
        fit = np.polyfit(freqs_new, fluxes_new, len(spectralIndex-1)).tolist()
        fit.reverse()
        return fit[0], fit[1:]
    else:
        return fluxes_new[0], None


def attenuate(beamMS, fluxes, RADeg, DecDeg, spectralIndex=None, referenceFrequency=None,
              timeIndx=0.5, invert=False):
    """
    Returns flux attenuated by primary beam.

    Note: the attenuation is approximated using the array factor beam from the
    first station in the beam MS only (and it is assumed that this station is at
    LOFAR core). This approximation has been found to produce reasonable results
    for a typical LOFAR observation but may not work well for atypical observations.

    Parameters
    ----------
    beamMS : str
        Measurement set for which the beam model is made
    fluxes : list
        List of fluxes to attenuate
    RADeg : list
        List of RA values in degrees
    DecDeg : list
        List of Dec values in degrees
    spectralIndex : list, optional
        List of spectral indices to adjust
    referenceFrequency : list, optional
        List of reference frequency of polynomial fit
    timeIndx : float (between 0 and 1), optional
        Time as fraction of that covered by the beamMS for which the beam is
        calculated
    invert : bool, optional
        If True, invert the beam (i.e. to un-attenuate the flux)

    Returns
    -------
    attFluxes : numpy array
        Attenuated fluxes
    adjSpectralIndex : numpy array
        Adjusted spectral indices. Returned only if spectralIndex is not None

    """
    import numpy as np
    import pyrap.tables as pt
    import multiprocessing
    import itertools

    t = pt.table(beamMS, ack=False)
    time = None
    ant1 = -1
    ant2 = 1
    while time is None:
        ant1 += 1
        ant2 += 1
        tt = t.query('ANTENNA1=={0} AND ANTENNA2=={1}'.format(ant1, ant2), columns='TIME')
        time = tt.getcol("TIME")
    t.close()
    if timeIndx < 0.0:
        timeIndx = 0.0
    if timeIndx > 1.0:
        timeIndx = 1.0
    time = min(time) + (max(time) - min(time)) * timeIndx
    sw = pt.table(beamMS+'::SPECTRAL_WINDOW', ack=False)
    numchannels = sw.col('NUM_CHAN')[0]
    startfreq = np.min(sw.col('CHAN_FREQ')[0])
    channelwidth = sw.col('CHAN_WIDTH')[0][0]
    sw.close()

    attFluxes = []
    adjSpectralIndex = []
    if type(fluxes) is not list:
        fluxes = list(fluxes)
    if type(RADeg) is not list:
        RADeg = list(RADeg)
    if type(DecDeg) is not list:
        DecDeg = list(DecDeg)
    if spectralIndex is not None:
        spectralIndex_list = spectralIndex
    else:
        spectralIndex_list = [None] * len(fluxes)
    if referenceFrequency is not None:
        referenceFrequency_list = referenceFrequency
    else:
        referenceFrequency_list = [None] * len(fluxes)

    pool = multiprocessing.Pool(multiprocessing.cpu_count())
    args_list = list(zip(RADeg, DecDeg, spectralIndex_list, fluxes, referenceFrequency_list, itertools.repeat(beamMS),
                         itertools.repeat(time), itertools.repeat(ant1), itertools.repeat(numchannels),
                         itertools.repeat(startfreq), itertools.repeat(channelwidth), itertools.repeat(invert)))
    results = pool.map(apply_beam_star, args_list)
    pool.close()
    pool.join()
    for fl, si in results:
        attFluxes.append(fl)
        adjSpectralIndex.append(si)

    if spectralIndex is not None:
        return np.array(attFluxes), np.array(adjSpectralIndex)
    else:
        return np.array(attFluxes)


def polynomial(stokesI, coeff, nu):
    result = stokesI
    for i in range(len(coeff)):
        result += coeff[i] * nu**(i+1)

    return result


def radec2xy(RA, Dec, refRA=None, refDec=None, crdelt=None):
    """
    Returns x, y for input ra, dec.

    Note that the reference RA and Dec must be the same in calls to both
    radec2xy() and xy2radec() if matched pairs of (x, y) <=> (RA, Dec) are
    desired.

    Parameters
    ----------
    RA : list
        List of RA values in degrees
    Dec : list
        List of Dec values in degrees
    refRA : float, optional
        Reference RA in degrees.
    refDec : float, optional
        Reference Dec in degrees
    crdelt: float, optional
        Delta in degrees for sky grid

    Returns
    -------
    x, y : list, list
        Lists of x and y pixel values corresponding to the input RA and Dec
        values

    """
    import numpy as np

    x = []
    y = []
    if refRA is None:
        refRA = RA[0]
    if refDec is None:
        refDec = Dec[0]

    # Make wcs object to handle transformation from ra and dec to pixel coords.
    w = makeWCS(refRA, refDec, crdelt=crdelt)

    for ra_deg, dec_deg in zip(RA, Dec):
        ra_dec = np.array([[ra_deg, dec_deg]])
        x.append(w.wcs_world2pix(ra_dec, 0)[0][0])
        y.append(w.wcs_world2pix(ra_dec, 0)[0][1])

    return x, y


def xy2radec(x, y, refRA=0.0, refDec=0.0, crdelt=None):
    """
    Returns x, y for input ra, dec.

    Note that the reference RA and Dec must be the same in calls to both
    radec2xy() and xy2radec() if matched pairs of (x, y) <=> (RA, Dec) are
    desired.

    Parameters
    ----------
    x : list
        List of x values in pixels
    y : list
        List of y values in pixels
    refRA : float, optional
        Reference RA in degrees
    refDec : float, optional
        Reference Dec in degrees
    crdelt: float, optional
        Delta in degrees for sky grid

    Returns
    -------
    RA, Dec : list, list
        Lists of RA and Dec values corresponding to the input x and y pixel
        values

    """
    import numpy as np

    RA = []
    Dec = []

    # Make wcs object to handle transformation from ra and dec to pixel coords.
    w = makeWCS(refRA, refDec, crdelt=crdelt)

    for xp, yp in zip(x, y):
        x_y = np.array([[xp, yp]])
        RA.append(w.wcs_pix2world(x_y, 0)[0][0])
        Dec.append(w.wcs_pix2world(x_y, 0)[0][1])

    return RA, Dec


def makeWCS(refRA, refDec, crdelt=None):
    """
    Makes simple WCS object.

    Parameters
    ----------
    refRA : float
        Reference RA in degrees
    refDec : float
        Reference Dec in degrees
    crdelt: float, optional
        Delta in degrees for sky grid

    Returns
    -------
    w : astropy.wcs.WCS object
        A simple TAN-projection WCS object for specified reference position

    """
    from astropy.wcs import WCS
    import numpy as np

    w = WCS(naxis=2)
    w.wcs.crpix = [1000, 1000]
    if crdelt is None:
        crdelt = 0.066667  # 4 arcmin
    w.wcs.cdelt = np.array([-crdelt, crdelt])
    w.wcs.crval = [refRA, refDec]
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    w.wcs.set_pv([(2, 1, 45.0)])

    return w


def matchSky(LSM1, LSM2, radius=0.1, byPatch=False, nearestOnly=False):
    """
    Matches two sky models by position.

    Parameters
    ----------
    LSM1 : SkyModel object
        Sky model for which match indices are desired
    LSM2 : SkyModel object
        Sky model to match against
    radius : float or str, optional
        Radius in degrees (if float) or 'value unit' (if str; e.g., '30 arcsec')
        for matching when matchBy='position'
    byPatch : bool, optional
        If True, matching is done by patches
    nearestOnly : bool, optional
        If True, only the nearest of multiple matches is returned

    Returns
    -------
    matches1, matches2 : np.array, np.array
        matches1 is the array of indices of LSM1 that have matches in LSM2
        within the specified radius. matches2 is the array of indices of LSM2
        for the same sources.

    """

    from astropy.coordinates import SkyCoord, Angle
    from astropy import units as u
    from distutils.version import LooseVersion
    import numpy as np
    import scipy

    log = logging.getLogger('LSMTool')
    if LooseVersion(scipy.__version__) < LooseVersion('0.11.0'):
        log.debug('The installed version of SciPy contains a bug that affects catalog matching. '
                  'Falling back on (slower) matching script.')
        from operations._matching import match_coordinates_sky
    else:
        from astropy.coordinates.matching import match_coordinates_sky

    if byPatch:
        RA, Dec = LSM1.getPatchPositions(asArray=True)
        catalog1 = SkyCoord(RA, Dec, unit=(u.degree, u.degree), frame='fk5')
        RA, Dec = LSM2.getPatchPositions(asArray=True)
        catalog2 = SkyCoord(RA, Dec, unit=(u.degree, u.degree), frame='fk5')
    else:
        catalog1 = SkyCoord(LSM1.getColValues('Ra'), LSM1.getColValues('Dec'),
                            unit=(u.degree, u.degree))
        catalog2 = SkyCoord(LSM2.getColValues('Ra'), LSM2.getColValues('Dec'),
                            unit=(u.degree, u.degree))
    idx, d2d, d3d = match_coordinates_sky(catalog1, catalog2)

    try:
        radius = float(radius)
    except ValueError:
        pass
    if type(radius) is float:
        radius = '{0} degree'.format(radius)
    radius = Angle(radius).degree
    matches1 = np.where(d2d.value <= radius)[0]
    matches2 = idx[matches1]

    if nearestOnly:
        filter = []
        for i in range(len(matches1)):
            mind = np.where(matches2 == matches2[i])[0]
            nMatches = len(mind)
            if nMatches > 1:
                mradii = d2d.value[matches1][mind]
                if d2d.value[matches1][i] == np.min(mradii):
                    filter.append(i)
            else:
                filter.append(i)
        matches1 = matches1[filter]
        matches2 = matches2[filter]

    return matches1, matches2


def calculateSeparation(ra1, dec1, ra2, dec2):
    """
    Returns angular separation between two coordinates (all in degrees).

    Parameters
    ----------
    ra1 : float or numpy array
        RA of coordinate 1 in degrees
    dec1 : float or numpy array
        Dec of coordinate 1 in degrees
    ra2 : float
        RA of coordinate 2 in degrees
    dec2 : float
        Dec of coordinate 2 in degrees

    Returns
    -------
    separation : astropy Angle or numpy array
        Angular separation in degrees

    """
    from astropy.coordinates import SkyCoord
    import astropy.units as u

    coord1 = SkyCoord(ra1, dec1, unit=(u.degree, u.degree), frame='fk5')
    coord2 = SkyCoord(ra2, dec2, unit=(u.degree, u.degree), frame='fk5')

    return coord1.separation(coord2)


def getFluxAtSingleFrequency(LSM, targetFreq=None, aggregate=None):
    """
    Returns flux density at given target frequency, adjusted according to the
    spectral index.

    Parameters
    ----------
    LSM : sky model object
        RA of coordinate 1 in degrees
    targetFreq : float, optional
        Frequency in Hz. If None, the median is used
    aggregate : str, optional
        Aggregation to use

    Returns
    -------
    fluxes : numpy array
        Flux densities in Jy

    """
    import numpy as np

    # Calculate flux densities
    if targetFreq is None:
        if 'ReferenceFrequency' in LSM.getColNames():
            refFreq = LSM.getColValues('ReferenceFrequency', aggregate=aggregate)
        else:
            refFreq = np.array([LSM.table.meta['ReferenceFrequency']]*len(LSM))
        targetFreq = np.median(refFreq)
    fluxes = LSM.getColValues('I', aggregate=aggregate)

    try:
        alphas = LSM.getColValues('SpectralIndex', aggregate=aggregate).squeeze(axis=0)
    except (IndexError, ValueError):
        alphas = np.array([-0.8]*len(fluxes))
    nterms = alphas.shape[1]

    if 'LogarithmicSI' in LSM.table.meta:
        logSI = LSM.table.meta['LogarithmicSI']
    else:
        logSI = True

    if nterms > 1:
        for i in range(nterms):
            if logSI:
                fluxes *= 10.0**(alphas[:, i] * (np.log10(refFreq / targetFreq))**(i+1))
            else:
                # stokesI + term0 (nu/refnu - 1) + term1 (nu/refnu - 1)^2 + ...
                fluxes += alphas[:, i] * ((refFreq / targetFreq) - 1.0)**(i+1)
    else:
        if logSI:
            fluxes *= 10.0**(alphas * np.log10(refFreq / targetFreq))
        else:
            fluxes += alphas * ((refFreq / targetFreq) - 1.0)

    return fluxes


def make_template_image(image_name, reference_ra_deg, reference_dec_deg, reference_freq,
                        ximsize=512, yimsize=512, cellsize_deg=0.000417, fill_val=0):
    """
    Make a blank image and save it to disk

    Parameters
    ----------
    image_name : str
        Filename of output image
    reference_ra_deg : float
        RA for center of output image
    reference_dec_deg : float
        Dec for center of output image
    reference_freq  : float
        Ref freq of output image
    ximsize : int, optional
        Size of output image
    yimsize : int, optional
        Size of output image
    cellsize_deg : float, optional
        Size of a pixel in degrees
    fill_val : int, optional
        Value with which to fill the image
    """
    import numpy as np
    from astropy.io import fits as pyfits

    # Make fits hdu
    # Axis order is [STOKES, FREQ, DEC, RA]
    shape_out = [1, 1, yimsize, ximsize]
    hdu = pyfits.PrimaryHDU(np.ones(shape_out, dtype=np.float32)*fill_val)
    hdulist = pyfits.HDUList([hdu])
    header = hdulist[0].header

    # Add RA, Dec info
    i = 1
    header['CRVAL{}'.format(i)] = reference_ra_deg
    header['CDELT{}'.format(i)] = -cellsize_deg
    header['CRPIX{}'.format(i)] = ximsize / 2.0
    header['CUNIT{}'.format(i)] = 'deg'
    header['CTYPE{}'.format(i)] = 'RA---SIN'
    i += 1
    header['CRVAL{}'.format(i)] = reference_dec_deg
    header['CDELT{}'.format(i)] = cellsize_deg
    header['CRPIX{}'.format(i)] = yimsize / 2.0
    header['CUNIT{}'.format(i)] = 'deg'
    header['CTYPE{}'.format(i)] = 'DEC--SIN'
    i += 1

    # Add STOKES info
    header['CRVAL{}'.format(i)] = 1.0
    header['CDELT{}'.format(i)] = 1.0
    header['CRPIX{}'.format(i)] = 1.0
    header['CUNIT{}'.format(i)] = ''
    header['CTYPE{}'.format(i)] = 'STOKES'
    i += 1

    # Add frequency info
    del_freq = 1e8
    header['RESTFRQ'] = reference_freq
    header['CRVAL{}'.format(i)] = reference_freq
    header['CDELT{}'.format(i)] = del_freq
    header['CRPIX{}'.format(i)] = 1.0
    header['CUNIT{}'.format(i)] = 'Hz'
    header['CTYPE{}'.format(i)] = 'FREQ'
    i += 1

    # Add equinox
    header['EQUINOX'] = 2000.0

    # Add telescope
    header['TELESCOP'] = 'LOFAR'

    hdulist[0].header = header
    hdulist.writeto(image_name, overwrite=True)
    hdulist.close()


def gaussian_fcn(g, x1, x2, const=False):
    """
    Evaluate a Gaussian on the given grid.

    Parameters
    ----------
    g: list
        List of Gaussian parameters:
        [peak_flux, xcen, ycen, FWHMmaj, FWHMmin, PA_E_of_N]
    x1, x2: grid (as produced by numpy.mgrid)
        Grid coordinates on which to evaluate the Gaussian
    const : bool, optional
        If True, all values are set to the peak_flux

    Returns
    -------
    img : array
        Image of Gaussian
    """
    from math import radians, sin, cos
    import numpy as np

    A, C1, C2, S1, S2, Th = g
    fwsig = 2.35482  # FWHM = fwsig * sigma
    S1 = S1 / fwsig
    S2 = S2 / fwsig
    Th = 360.0 - Th  # theta is angle E of N (ccw from +y axis)
    th = radians(Th)
    cs = cos(th)
    sn = sin(th)
    f1 = ((x1-C1)*cs + (x2-C2)*sn)/S1
    f2 = (-(x1-C1)*sn + (x2-C2)*cs)/S2
    gimg = A * np.exp(-(f1*f1 + f2*f2)/2)

    if const:
        mask = np.where(gimg/A > 1e-5)
        cimg = np.zeros(x1.shape)
        cimg[mask] = A
        return cimg
    else:
        return gimg


def tessellate(x_pix, y_pix, w, dist_pix):
    """
    Returns Voronoi tessellation vertices

    Parameters
    ----------
    x_pix : array
        Array of x pixel values for tessellation centers
    y_pix : array
        Array of y pixel values for tessellation centers
    w : WCS object
        WCS for transformation from pix to world coordinates
    dist_pix : float
        Distance in pixels from center to outer boundary of facets

    Returns
    -------
    verts : list
        List of facet vertices in (RA, Dec)
    """
    import numpy as np
    from scipy.spatial import Voronoi
    import shapely.geometry
    import shapely.ops

    # Get x, y coords for directions in pixels. We use the input calibration sky
    # model for this, as the patch positions written to the h5parm file by DPPP may
    # be different
    xy = []
    for RAvert, Decvert in zip(x_pix, y_pix):
        xy.append((RAvert, Decvert))

    # Generate array of outer points used to constrain the facets
    nouter = 64
    means = np.ones((nouter, 2)) * np.array(xy).mean(axis=0)
    offsets = []
    angles = [np.pi/(nouter/2.0)*i for i in range(0, nouter)]
    for ang in angles:
        offsets.append([np.cos(ang), np.sin(ang)])
    scale_offsets = dist_pix * np.array(offsets)
    outer_box = means + scale_offsets

    # Tessellate and clip
    points_all = np.vstack([xy, outer_box])
    vor = Voronoi(points_all)
    lines = [
        shapely.geometry.LineString(vor.vertices[line])
        for line in vor.ridge_vertices
        if -1 not in line
    ]
    polygons = [poly for poly in shapely.ops.polygonize(lines)]
    verts = []
    for poly in polygons:
        verts_xy = poly.exterior.xy
        verts_deg = []
        for x, y in zip(verts_xy[0], verts_xy[1]):
            x_y = np.array([[y, x, 0.0, 0.0]])
            ra_deg, dec_deg = w.wcs_pix2world(x_y, 0)[0][0], w.wcs_pix2world(x_y, 0)[0][1]
            verts_deg.append((ra_deg, dec_deg))
        verts.append(verts_deg)

    # Reorder to match the initial ordering
    ind = []
    for poly in polygons:
        for j, (xs, ys) in enumerate(zip(x_pix, y_pix)):
            if poly.contains(shapely.geometry.Point(xs, ys)):
                ind.append(j)
                break
    verts = [verts[i] for i in ind]

    return verts



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation implements merging of patches
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging

log = logging.getLogger('LSMTool.MERGE')
log.debug('Loading MERGE module.')


def run(step, parset, LSM):

    outFile = parset.getString('.'.join(["LSMTool.Steps", step, "OutFile"]), '' )
    patches = parset.getStringVector('.'.join(["LSMTool.Steps", step, "Patches"]), [] )
    name = parset.getString('.'.join(["LSMTool.Steps", step, "Name"]), '' )
    if name == '':
        name = None

    try:
        result = merge(LSM, patches, name)
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    # Write to outFile
    if outFile != '' and result == 0:
        LSM.write(outFile, clobber=True)

    return result


def merge(LSM, patches, name=None):
    """
    Merge two or more patches together

    Parameters
    ----------
    LSM : SkyModel object
        Input sky model
    patches : list of str
        List of patches to merge
    name : str, optional
        Name of resulting merged patch. If None, the merged patch uses the name
        of the first patch in the input patches list

    Examples
    --------
    Merge three patches into one patch named 'binmerged'::

        >>> LSM = lsmtool.load('sky.model')
        >>> merge(LSM, ['bin0', 'bin1', 'bin2'], 'binmerged')

    """
    if len(LSM) == 0:
        log.error('Sky model is empty.')
        return

    if name is None:
        name = patches[0]

    indices = []
    for patchName in patches:
        indices.append(LSM.getRowIndex(patchName))
    for patchName, ind in zip(patches, indices):
        if ind is None:
            raise ValueError("Could not find patch '{0}'.".format(patchName))
        else:
            LSM.table['Patch'][ind] = name
    LSM._updateGroups()
    LSM._addHistory("MERGE ('{0}' into '{1}')".format(patches, name))
    LSM._info()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation implements adding of sources to the sky model
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging

log = logging.getLogger('LSMTool.ADD')
log.debug('Loading ADD module.')


def run(step, parset, LSM):

    from ..tableio import allowedColumnNames

    outFile = parset.getString('.'.join(["LSMTool.Steps", step, "OutFile"]), '' )
    colNamesVals = {}
    for colName in allowedColumnNames:
        val = parset.getString('.'.join(["LSMTool.Steps",
            step, allowedColumnNames[colName]]), '' )
        if val != '':
            try:
                val = float(val)
            except ValueError:
                pass
            colNamesVals[allowedColumnNames[colName]] = val

    try:
        add(LSM, colNamesVals)
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    if outFile != '' and result == 0:
        LSM.write(outFile, clobber=True)

    return result


def add(LSM, colNamesVals):
    """
    Add a source to the sky model.

    Parameters
    ----------
    LSM : SkyModel object
        Input sky model
    colNamesVals : dict
        A dictionary that specifies the column values for the source to be added

    Examples:
    ---------
    Add a point source::

        >>> source = {'Name':'src1', 'Type':'POINT', 'Ra':'12:32:10.1',
            'Dec':'23.43.21.21', 'I':2.134}
        >>> add(LSM, source)

    """
    sourceNames = LSM.getColValues('Name').tolist()
    if colNamesVals['Name'] in sourceNames:
        raise ValueError('A source with the same name already exists.')

    LSM.setRowValues(colNamesVals)
    LSM._updateGroups()
    LSM._addHistory("ADD (source '{0}')".format(colNamesVals['Name']))
    LSM._info()



# -*- coding: utf-8 -*-
#
# Defines filter functions used by the remove and select operations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging
log = logging.getLogger('LSMTool.Filter')
try:
    dict.iteritems
except AttributeError:
    # Python 3
    def itervalues(d):
        return iter(d.values())
    def iteritems(d):
        return iter(d.items())
else:
    # Python 2
    def itervalues(d):
        return d.itervalues()
    def iteritems(d):
        return d.iteritems()


def filter(LSM, filterExpression, exclusive=False, aggregate=None,
    applyBeam=False, useRegEx=False, force=True):
    """
    Filters the sky model, keeping all sources that meet the given expression.

    After filtering, the sky model contains only those sources for which the
    given filter expression is true.

    Parameters
    ----------
    filterExpression : str, dict, list, or numpy array
        - If string:
            A string specifying the filter expression in the form:
            '<property> <operator> <value> [<units>]'
            (e.g., 'I <= 10.5 Jy').

        - If dict:
            The filter can also be given as a dictionary in the form:
            {'filterProp':property, 'filterOper':operator,
                'filterVal':value, 'filterUnits':units}

        - If list:
            The filter can also be given as a list of:
            [property, operator, value] or
            [property, operator, value, units]

        - If numpy array:
            The indices to filter on can be specified directly as a numpy array
            of row or patch indices such as:
            array([ 0,  2, 19, 20, 31, 37])

            or as a numpy array of bools with the same length as the sky model.

            If a numpy array is given and the indices correspond to patches, then
            set aggregate=True.

        The property to filter on must be one of the following:
            - a valid column name
            - the filename of a mask image

        Supported operators are:
            - !=
            - <=
            - >=
            - >
            - <
            - = (or '==')
        Units are optional and must be specified as required by astropy.units.
    exclusive : bool, optional
        If False, sources that meet the filter expression are kept. If True,
        sources that do not meet the filter expression are kept.
    aggregate : str or bool, optional
        If set, the values are aggregated over the patch members. The following
        aggregation functions are available:
            - 'sum': sum of patch values
            - 'mean': mean of patch values
            - 'wmean': Stokes I weighted mean of patch values
            - 'min': minimum of patch values
            - 'max': maximum of patch values
            - True: only valid when the filter indices are specified directly as
                a numpy array. If True, filtering is done on patches instead of
                sources.
    applyBeam : bool, optional
        If True, apparent fluxes will be used.
    useRegEx : bool, optional
        If True, string matching will use regular expression matching. If
        False, string matching uses Unix filename matching.
    force : bool, optional
        If True, filters that result in empty sky models are allowed. If
        False, such filters are not applied and the sky model is unaffected.

    Examples
    --------
    Filter on column 'I' (Stokes I flux). This filter will select all sources
    with Stokes I flux greater than 1.5 Jy::

        >>> LSM = lsmtool.load('sky.model')
        >>> filter(LSM, 'I > 1.5 Jy')
        INFO: Filtered out 1102 sources.

    If the sky model has patches and the filter is desired per patch, use
    ``aggregate = True``::

        >>> filter(LSM, 'I > 1.5 Jy', aggregate=True)

    Filter on source names, keeping those that match "src*_1?"::

        >>> filter(LSM, 'Name == src*_1?')

    Use a CASA clean mask image to keep sources that lie in masked regions::

        >>> filter(LSM, 'clean_mask.mask == True')

    Filter on patch size::

        >>> sizes = LSM.getPatchSizes(units='arcsec', weight=True)
        >>> indices = numpy.where(sizes <= maj_cut_arcsec)
        >>> filter(LSM, indices, aggregate=True)

    or more simply::

        >>> sizes = LSM.getPatchSizes(units='arcsec', weight=True)
        >>> filter(LSM, sizes < maj_cut_arcsec, aggregate=True)

    """
    import numpy as np
    import math as m
    import os
    from astropy.table import Table

    if len(LSM) == 0:
        log.error('Sky model is empty.')
        return

    if filterExpression is None:
        raise ValueError('No filter expression specified.')

    if LSM.hasPatches and aggregate is not None:
        nrows = len(LSM.getPatchNames())
    else:
        nrows = len(LSM)

    filt = None
    if type(filterExpression) is list:
        try:
            if len(filterExpression) == 3:
                filterProp, filterOperStr, filterVal = filterExpression
                filterUnits = None
                filterOper, f = convertOperStr(filterOperStr)
            elif len(filterExpression) == 4:
                filterProp, filterOperStr, filterVal, filterUnits = filterExpression
                filterOper, f = convertOperStr(filterOperStr)
        except Exception:
            raise ValueError("Could not parse filter.")

    elif type(filterExpression) is dict:
        if ('filterProp' in filterExpression.keys() and
            'filterOper' in filterExpression.keys() and
            'filterVal' in filterExpression.keys()):
            filterProp = filterExpression['filterProp']
            filterOperStr = filterExpression['filterOper']
            filterOper, f = convertOperStr(filterOperStr)
            filterVal = filterExpression['filterVal']
        else:
            raise ValueError("Could not parse filter.")
        if 'filterUnits' in filterExpression.keys():
            filterUnits = filterExpression['filterUnits']
        else:
            filterUnits = None

    elif type(filterExpression) is str:
        # Parse the filter expression
        filterProp, filterOper, filterVal, filterUnits = parseFilter(filterExpression)

    elif type(filterExpression) is np.ndarray:
        # Array of indices / bools
        if np.result_type(filterExpression) == 'bool':
            if len(filterExpression) == nrows:
                filt = [i for i in range(len(filterExpression)) if
                    filterExpression[i]]
            else:
                raise ValueError("Boolean filter arrays be of same length as "
                    "the sky model.")
        else:
            filt = filterExpression.tolist()

    else:
        raise ValueError("Could not parse filter.")

    if filt is None and filterProp is None:
        raise ValueError('Filter expression not understood')

    if filt is None:
        # Get the column values to filter on
        if LSM._verifyColName(filterProp, quiet=True) in LSM.table.colnames:
            filterProp = LSM._verifyColName(filterProp)
            colVals = LSM.getColValues(filterProp, units=filterUnits,
                aggregate=aggregate, applyBeam=applyBeam)
        else:
            # Assume filterProp is a mask filename and try to load mask
            if os.path.exists(filterProp):
                mask = filterProp
                RARad = LSM.getColValues('Ra', units='radian')
                DecRad = LSM.getColValues('Dec', units='radian')
                colVals = getMaskValues(mask, RARad, DecRad)
            else:
                raise ValueError('Could not parse filter.')

        # Do the filtering
        if colVals is None:
            raise ValueError("Could not parse filter.")
        filt = getFilterIndices(colVals, filterOper, filterVal, useRegEx=useRegEx)

    if exclusive:
        filt = [i for i in range(nrows) if i not in filt]
    if len(filt) == 0:
        if force:
            LSM.table.remove_rows(range(len(LSM.table)))
            LSM.hasPatches = False
            if exclusive:
                log.info('Removed all sources.')
                return
            else:
                log.info('Kept zero sources.')
                return
        else:
            raise RuntimeError('Filter would result in an empty sky model. '
                'Use force=True to override.')
    if len(filt) == len(LSM):
        if exclusive:
            log.info('Removed zero sources.')
        else:
            log.info('Kept all sources.')
        return

    if LSM.hasPatches and aggregate is not None:
        sourcesToKeep = LSM.getPatchNames()[filt]
        allsources = LSM.getColValues('Patch')
        indicesToKeep = [i for i in range(len(LSM)) if allsources[i] in sourcesToKeep]
        nPatchesOrig = len(LSM.table.groups)
        LSM.table = LSM.table[indicesToKeep]
        LSM._updateGroups()
        nPatchesNew = len(LSM.table.groups)
        if nPatchesOrig - nPatchesNew == 1:
            plustr = ''
        else:
            plustr = 'es'
        if exclusive:
            log.info('Removed {0} patch{1}.'.format(nPatchesOrig-nPatchesNew, plustr))
        else:
            if nPatchesNew > 1:
                plustr = 'es'
            log.info('Kept {0} patch{1}.'.format(nPatchesNew, plustr))
    else:
        nRowsOrig = len(LSM)
        LSM.table = LSM.table[filt]
        nRowsNew = len(LSM)
        if LSM.hasPatches:
            LSM._updateGroups()
        if exclusive:
            if nRowsOrig - nRowsNew == 1:
                plustr = ''
            else:
                plustr = 's'
            log.info('Removed {0} source{1}.'.format(nRowsOrig-nRowsNew, plustr))
        else:
            if nRowsNew == 1:
                plustr = ''
            else:
                plustr = 's'
            log.info('Kept {0} source{1}.'.format(nRowsNew, plustr))

    if type(filterExpression) is np.ndarray:
        history = 'with array of indices/bools'
    else:
        _, filterOperStr = convertOperStr(filterOper)
        history = '{0} {1} {2} {3}'.format(filterProp, filterOperStr, filterVal, filterUnits)
    LSM._addHistory('FILTER ({0})'.format(history))
    LSM._info()


def parseFilter(filterExpression):
    """
    Takes a filter expression and returns tuple of (property, operation, val,
    units), all as strings.
    """
    try:
        from ..tableio import allowedColumnNames
        from ..tableio import allowedColumnDefaults
    except:
        from .tableio import allowedColumnNames
        from .tableio import allowedColumnDefaults
    from itertools import groupby

    # Get operator function
    filterOper, filterOperStr = convertOperStr(filterExpression)
    if filterOper is None:
        return (None, None, None, None)

    filterParts = filterExpression.split(filterOperStr)
    if len(filterParts) != 2:
        raise ValueError("Filter expression must be of the form '<property> "
            "<operator> <value> <unit>'\nE.g., 'Flux >= 10 Jy'")

    # Get the column to filter on
    filterProp = filterParts[0].strip().lower()
    if filterProp not in allowedColumnNames:
        log.warn('"{0}" is not a valid column. Trying it as a mask '
            'filename instead...'.format(filterProp))
        filterProp = filterParts[0].strip() # don't use lower-case version if filename
        filterVal = filterParts[1].strip()
        return (filterProp, filterOper, input2bool(filterVal), None)

    # Get the filter value(s)
    filterValAndUnits = filterParts[1].strip()
    if allowedColumnDefaults[filterProp] == 'N/A':
        # Column values are strings. Allow only '==' and '!=' operators
        if filterOperStr not in ['=', '==', '!=']:
            raise ValueError("Filter operator '{0}' not allow with string columns. "
                "Supported operators are '!=' or '=' (or '==')".format(filterOperStr))

        # Check for a list of values
        if '[' in filterValAndUnits and ']' in filterValAndUnits:
            filterVal = filterValAndUnits.split(']')[0].strip('[')
            filterValParts = filterVal.split(',')
            filterVal = []
            for val in filterValParts:
                val = val.strip()
                val = val.strip('"')
                val = val.strip("'")
                filterVal.append(val)
        else:
            filterVal = filterValAndUnits.split(' ')[0].strip()
            filterVal = filterVal.strip('"')
            filterVal = filterVal.strip("'")
    else:
        # The column to filter is type float
        try:
            parts = [''.join(g).strip() for _, g in groupby(filterValAndUnits,
                str.isalpha)]
            if len(parts) > 1:
                if type(parts[1]) is str:
                    if parts[1].lower() == 'e':
                        # Check if number uses exponential notation (e.g., 1e8)
                        parts = [parts[0] + parts[1] + parts[2]] + parts[3:]
            filterVal = float(parts[0])
        except ValueError:
            raise ValueError('Filter value not understood.')

    # Try and get the units (only if filterVal is not a string)
    if type(filterVal) is str:
        filterUnits = None
    else:
        try:
            filterUnits = filterValAndUnits.split(']')
            if len(filterUnits) == 1:
                parts = [''.join(g).strip() for _, g in groupby(filterUnits[0],
                    str.isalpha)]
                if type(parts[1]) is str:
                    if parts[1].lower() == 'e':
                        # Check if number uses exponential notation (e.g., 1e8)
                        parts = [parts[0] + parts[1] + parts[2]] + parts[3:]
                filterUnits = parts[1]
            else:
                filterUnits = filterUnits[1].strip()
        except IndexError:
            filterUnits = None
        if filterUnits == '':
            filterUnits = None

    return (filterProp, filterOper, filterVal, filterUnits)


def input2bool(invar):
    if isinstance(invar, bool):
        return invar
    elif isinstance(invar, str):
        if invar.upper() == 'TRUE' or invar == '1':
            return True
        elif invar.upper() == 'FALSE' or invar == '0':
            return False
        else:
            raise ValueError('Cannot convert string "'+invar+'" to boolean!')
    elif isinstance(invar, int) or isinstance(invar, float):
        return bool(invar)
    else:
        raise TypeError('Unsupported data type:'+str(type(invar)))


def convertOperStr(operStr):
    """
    Returns operator function corresponding to string.
    """
    import operator as op

    filterOperStr = None
    ops = {'!=':op.ne, '<=':op.le, '>=':op.ge, '>':op.gt, '<':op.lt,
        '==':op.eq, '=':op.eq}

    if type(operStr) is str:
        for op in ops:
            if op in operStr:
                if filterOperStr is None:
                    filterOperStr = op
                elif len(op) > len(filterOperStr):
                    # Pick longer match
                    filterOperStr = op
    else:
        for k, v in iteritems(ops):
            if v == operStr:
                filterOperStr = k
    if filterOperStr is None:
        return None, None

    return ops[filterOperStr], filterOperStr


def getFilterIndices(colVals, filterOper, filterVal, useRegEx=False):
    """
    Returns the indices that correspond to the input filter expression.
    """
    import operator as op
    import fnmatch
    import re

    if type(filterVal) is not list:
        filterVal = [filterVal]
    filterInd = []
    for val in filterVal:
        if type(val) is str:
            # String -> use regular expression or Unix filename matching search
            if filterOper is op.eq:
                if useRegEx:
                    filt = [i for i, item in enumerate(colVals) if re.search(val, item) is not None]
                else:
                    filt = [i for i, item in enumerate(colVals) if fnmatch.fnmatch(item, val)]
            elif filterOper is op.ne:
                if useRegEx:
                    filt = [i for i, item in enumerate(colVals) if re.search(val, item) is None]
                else:
                    filt = [i for i, item in enumerate(colVals) if not fnmatch.fnmatch(item, val)]
            else:
                raise ValueError("Filter operator '{0}' not allow with string columns. "
                    "Supported operators are '!=' or '=' (or '==')".format(filterOper))
        else:
            filtBool = filterOper(colVals, val)
            filt = [f for f in range(len(colVals)) if filtBool[f]]
        filterInd += filt

    return filterInd


def getMaskValues(mask, RARad, DecRad):
    """
    Returns an array of mask values for each (RA, Dec) pair in radians.
    """
    import math
    import pyrap.images as pim
    import numpy as np

    maskdata = pim.image(mask)
    maskval = maskdata.getdata()[0][0]

    vals = []
    for raRad, decRad in zip(RARad, DecRad):
        (a, b, _, _) = maskdata.toworld([0, 0, 0, 0])
        (_, _, pixY, pixX) = maskdata.topixel([a, b, decRad, raRad])
        try:
            if maskval[int(pixY), int(pixX)] and pixX >= 0 and pixY >=0:
                vals.append(True)
            else:
                vals.append(False)
        except IndexError:
            vals.append(False)

    return np.array(vals)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module contains functions for matching coordinate catalogs.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import numpy as np

from astropy.extern import six

__all__ = ['match_coordinates_3d', 'match_coordinates_sky']

def match_coordinates_3d(matchcoord, catalogcoord, nthneighbor=1, storekdtree=True):
    """
    Finds the nearest 3-dimensional matches of a coordinate or coordinates in
    a set of catalog coordinates.

    This finds the 3-dimensional closest neighbor, which is only different
    from the on-sky distance if `distance` is set in either `matchcoord`
    or `catalogcoord`.

    Parameters
    ----------
    matchcoord : `~astropy.coordinates.SphericalCoordinatesBase`
        The coordinate(s) to match to the catalog.
    catalogcoord : `~astropy.coordinates.SphericalCoordinatesBase`
        The base catalog in which to search for matches. Typically this will
        be a coordinate object that is an array (i.e.,
        ``catalogcoord.isscalar == False``)
    nthneighbor : int, optional
        Which closest neighbor to search for.  Typically ``1`` is desired here,
        as that is correct for matching one set of coordinates to another.
        The next likely use case is ``2``, for matching a coordinate catalog
        against *itself* (``1`` is inappropriate because each point will find
        itself as the closest match).
    storekdtree : bool or str, optional
        If True or a string, will store the KD-Tree used for the computation
        in the `catalogcoord`.  This dramatically speeds up subsequent calls
        with the same catalog. If a str, it specifies the attribute name for
        `catalogcoord` that should store the KD-tree.

    Returns
    -------
    idx : integer array
        Indecies into `catalogcoord` to get the matched points for each
        `matchcoord`. Shape matches `matchcoord`.
    sep2d : `~astropy.units.quantity.Angle`
        The on-sky separation between the closest match for each `matchcoord` and
        the `matchcoord`. Shape matches `matchcoord`.
    dist3d : `~astropy.units.quantity.Quantity`
        The 3D distance between the closest match for each `matchcoord` and
        the `matchcoord`. Shape matches `matchcoord`.

    Notes
    -----
    This function requires `scipy` to be installed or it will fail.
    """
    from warnings import warn
    from ._kdtree import KDTree

    if storekdtree is True:
        storekdtree = '_kdtree'

    # figure out where any cached KDTree might be
    if isinstance(storekdtree, six.string_types):
        kdt = getattr(catalogcoord, storekdtree, None)
        if kdt is not None and not isinstance(kdt, KDTree):
            raise ValueError('Invalid `storekdtree` string:' + storekdtree)
    elif isinstance(storekdtree, KDTree):
        kdt = storekdtree
        storekdtree = None
    elif not storekdtree:
        kdt = None
    else:
        raise ValueError('Invalid `storekdtree` argument:' +
                          str(storekdtree))

    if kdt is None:
        #need to build the cartesian KD-tree for the catalog
        cart = catalogcoord.cartesian
        flatxyz = cart.reshape((3, np.prod(cart.shape) // 3))
        kdt = KDTree(flatxyz.value.T)

    #make sure coordinate systems match
    matchcoord = matchcoord.transform_to(catalogcoord.__class__)

    #make sure units match
    catunit = catalogcoord.cartesian.unit
    cart = matchcoord.cartesian.to(catunit)

    flatxyz = cart.reshape((3, np.prod(cart.shape) // 3))
    dist, idx = kdt.query(flatxyz.T, nthneighbor)

    if nthneighbor > 1:  # query gives 1D arrays if k=1, 2D arrays otherwise
        dist = dist[:, -1]
        idx = idx[:, -1]

    if storekdtree:
        #cache the kdtree
        setattr(catalogcoord, storekdtree, kdt)

    sep2d = catalogcoord[idx].separation(matchcoord)
    return idx.reshape(cart.shape[1:]), sep2d, dist.reshape(cart.shape[1:]) * catunit


def match_coordinates_sky(matchcoord, catalogcoord, nthneighbor=1, storekdtree=True):
    """
    Finds the nearest on-sky matches of a coordinate or coordinates in
    a set of catalog coordinates.

    This finds the on-sky closest neighbor, which is only different from the
    3-dimensional match if `distance` is set in either `matchcoord`
    or `catalogcoord`.

    Parameters
    ----------
    matchcoord : `~astropy.coordinates.SphericalCoordinatesBase`
        The coordinate(s) to match to the catalog.
    catalogcoord : `~astropy.coordinates.SphericalCoordinatesBase`
        The base catalog in which to search for matches. Typically this will
        be a coordinate object that is an array (i.e.,
        ``catalogcoord.isscalar == False``)
    nthneighbor : int, optional
        Which closest neighbor to search for.  Typically ``1`` is desired here,
        as that is correct for matching one set of coordinates to another.
        The next likely use case is ``2``, for matching a coordinate catalog
        against *itself* (``1`` is inappropriate because each point will find
        itself as the closest match).
    storekdtree : bool or str, optional
        If True or a string, will store the KD-Tree used for the computation
        in the `catalogcoord`.  This dramatically speeds up subsequent calls
        with the same catalog. If a str, it specifies the attribute name for
        `catalogcoord` that should store the KD-tree.

    Returns
    -------
    idx : integer array
        Indecies into `catalogcoord` to get the matched points for each
        `matchcoord`. Shape matches `matchcoord`.
    sep2d : `~astropy.units.quantity.Angle`
        The on-sky separation between the closest match for each `matchcoord` and
        the `matchcoord`. Shape matches `matchcoord`.
    dist3d : `~astropy.units.quantity.Quantity`
        The 3D distance between the closest match for each `matchcoord` and
        the `matchcoord`. Shape matches `matchcoord`.

    Notes
    -----
    This function requires `scipy` to be installed or it will fail.
    """
    dcoo = matchcoord._distance
    cpcoo = matchcoord._cartpoint
    dcat = catalogcoord._distance
    cpcat = catalogcoord._cartpoint
    try:
        matchcoord._distance = matchcoord._cartpoint = None
        catalogcoord._distance = catalogcoord._cartpoint = None
        return match_coordinates_3d(matchcoord, catalogcoord, nthneighbor, storekdtree)
    finally:
        matchcoord._distance = dcoo
        matchcoord._cartpoint = cpcoo
        catalogcoord._distance = dcat
        catalogcoord._cartpoint = cpcat



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation implements grouping of sources into patches
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging

log = logging.getLogger('LSMTool.MOVE')
log.debug('Loading MOVE module.')


def run(step, parset, LSM):

    outFile = parset.getString('.'.join(["LSMTool.Steps", step, "OutFile"]), '' )
    name = parset.getString('.'.join(["LSMTool.Steps", step, "Name"]), '' )
    position = parset.getStringVector('.'.join(["LSMTool.Steps", step, "Position"]), [] )
    shift = parset.getStringVector('.'.join(["LSMTool.Steps", step, "Shift"]), [] )

    if len(position) < 2:
        position = None
    if len(shift) < 2:
        shift = None

    try:
        move(LSM, name, position, shift)
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    # Write to outFile
    if outFile != '' and result == 0:
        LSM.write(outFile, clobber=True)

    return result


def move(LSM, name, position=None, shift=None, xyshift=None, fitsFile=None):
    """
    Move or shift a source or sources.

    If both a position and a shift are specified, a source is moved to the
    new position and then shifted. Note that only a single source can be
    moved to a new position. However, multiple sources can be shifted.

    If an xyshift is specified, a FITS file must also be specified to define
    the WCS system. If a position, a shift, and an xyshift are all specified,
    a source is moved to the new position, shifted in RA and Dec, and then
    shifted in x and y.

    Parameters
    ----------
    LSM : SkyModel object
        Input sky model
    name : str or list
        Source name or list of names (can include wildcards)
    position : list, optional
        A list specifying a new position as [RA, Dec] in either makesourcedb
        format (e.g., ['12:23:43.21', '+22.34.21.2']) or in degrees (e.g.,
        [123.2312, 23.3422])
    shift : list, optional
        A list specifying the shift as [RAShift, DecShift] in degrees (e.g.,
        [0.02312, 0.00342])
    xyshift : list, optional
        A list specifying the shift as [xShift, yShift] in pixels. A FITS file
        must be specified with the fitsFILE argument
    fitsFile : str, optional
        A FITS file from which to take WCS information to transform the pixel
        coordinates to RA and Dec values. The xyshift argument must be specfied
        for this to be useful

    Examples
    --------
    Move source '1609.6+6556' to a new position::

        >>> LSM = lsmtool.load('sky.model')
        >>> move(LSM, '1609.6+6556', position=['16:10:00', '+65.57.00'])

    Shift the source by 10 arcsec in Dec::

        >>> move(LSM, '1609.6+6556', shift=[0.0, 10.0/3600.0])

    Shift all sources by 10 pixels in x::

        >>> move(LSM, '*', xyshift=[10, 0], fitsFile='image.fits')

    """
    from .. import tableio
    from astropy.io.fits import getheader
    from astropy import wcs

    if len(LSM) == 0:
        log.error('Sky model is empty.')
        return

    if fitsFile is not None:
        if xyshift is None:
            log.warn("A FITS file is specified, but no xyshift is specified.")
        hdr = getheader(fitsFile, 1)
        w = wcs.WCS(hdr)
    elif xyshift is not None:
        raise ValueError("A FITS file must be specified to use xyshift.")
    if position is None and shift is None and xyshift is None:
        raise ValueError("One of positon, shift, or xyshift must be specified.")

    sourceNames = LSM.getColValues('Name')
    table = LSM.table.copy()
    indx = LSM._getNameIndx(name)
    if indx is not None:
        if position is not None:
            if len(indx) > 1:
                raise ValueError('Only one source can be moved to a new position')
            try:
                table['Ra'][indx] = tableio.RA2Angle(position[0])[0]
                table['Dec'][indx] = tableio.Dec2Angle(position[1])[0]
            except Exception as e:
                raise ValueError('Could not parse position: {0}'.format(e))
        if shift is not None:
            for ind in indx:
                RA = LSM.table['Ra'][ind] + tableio.RA2Angle(shift[0])
                Dec = LSM.table['Dec'][ind] + tableio.Dec2Angle(shift[1])
                table['Ra'][ind] = tableio.RA2Angle(RA)[0]
                table['Dec'][ind] = tableio.Dec2Angle(Dec)[0]
        if xyshift is not None:
            for ind in indx:
                radec = np.array([LSM.table['Ra'][ind], LSM.table['Dec'][ind]])
                xy = w.wcs_world2pix(radec, 1)
                xy[0] += xyshift[0]
                xy[1] += xyshift[1]
                RADec = w.wcs_pix2world(xy, 1)
                table['Ra'][ind] = tableio.RA2Angle(RADec[0])[0]
                table['Dec'][ind] = tableio.Dec2Angle(RADec[1])[0]
        LSM.table = table
    elif LSM.hasPatches:
        indx = LSM._getNameIndx(name, patch=True)
        patchNames = LSM.getPatchNames()
        if indx is not None:
            if position is not None:
                if len(indx) > 1:
                    raise ValueError('Only one source can be moved to a new position')
                try:
                    position[0] = tableio.RA2Angle(position[0])[0]
                    position[1] = tableio.Dec2Angle(position[1])[0]
                except Exception as e:
                    raise ValueError('Could not parse position: {0}'.format(e))
                table.meta[name] = position
            if shift is not None:
                for ind in indx:
                    pname = patchNames[ind]
                    position = LSM.table.meta[name]
                    table.meta[name] = [position[0] + tableio.RA2Angle(shift[0]),
                        position[1] + tableio.Dec2Angle(shift[1])]
            if xyshift is not None:
                for ind in indx:
                    pname = patchNames[ind]
                    radec = np.array(LSM.table.meta[name])
                    xy = w.wcs_world2pix(radec, 1)
                    xy[0] += xyshift[0]
                    xy[1] += xyshift[1]
                    RADec = w.wcs_pix2world(xy, 1)
                    table.meta[name] = [RADec[0], RADec[1]]
            LSM.table = table
        else:
            raise ValueError("Could not find patch '{0}'.".format(name))
    else:
        raise ValueError("Could not find source '{0}'.".format(name))

    history = "'{0}' by {1} deg".format(name, shift)
    if position is not None:
        history += ', moved to {0}'.format(position)
    if shift is not None:
        history += ', shifted by {1} degrees'.format(shift)
    if xyshift is not None:
        history += ', shifted by {1} pixels'.format(shift)
    LSM._addHistory("Move ({0})".format(history))



# -*- coding: utf-8 -*-
#
# Defines the tessellate functions used by the group operation
#
# Based on the IDL version of the Voronoi binning method of:
#   Cappellari M., Copin Y., 2003, MNRAS, 342, 345
# Modified by David Rafferty as required for integration into LSMTool
#
# This software is provided as is without any warranty whatsoever.
# Permission to use, copy and distribute unmodified copies for
# non-commercial purposes is granted. Permission to modify for
# personal or internal use is granted, provided this copyright
# and disclaimer are included unchanged at the beginning of
# the file. All other rights are reserved.


from __future__ import print_function
import numpy as np
from numpy import sum, sqrt, min, max, any
from numpy import argmax, argmin, mean, abs
from numpy import int32 as Nint
from numpy import float32 as Nfloat
import copy


def get_skymodel(skymodel_fn):
    """Returns x, y, flux arrays for input skymodel
    """
    import math as m
    from astropy import WCS

    y = []
    x = []
    fluxes = []

    # Make wcs object to handle transformation from ra and dec to pixel coords.
    w = WCS(naxis=2)
    w.wcs.crpix = [-234.75, 8.3393]
    w.wcs.cdelt = np.array([-0.066667, 0.066667])
    w.wcs.crval = [0, -90]
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    w.wcs.set_pv([(2, 1, 45.0)])
    arcsec_per_pix = abs(w.wcs.cdelt[0]) * 3600.0 # arcsec/pixel

    for line in open(skymodel_fn):
        sline=line.split(',')
        if line.startswith("FORMAT"): continue
        if line[0] == '#': continue
        if line[0] == '\n': continue
        name = str(sline[0])
        srctype = str(sline[1])
        ra_src = str(sline[2]).split(':')
        ra_deg = float(ra_src[0])*15.0 + (float(ra_src[1])/60.0)*15.0 + (float(ra_src[2])
            /3600.0)*15.0
        dec_src = str(sline[3]).split('.')
        if len(dec_src) == 3:
            dec_deg = float(dec_src[0]) + (float(dec_src[1])/60.0) + (float(dec_src[2])/3600.0)
        else:
            dec_deg = float(dec_src[0]) + (float(dec_src[1])/60.0) + (float(dec_src[2]
                + '.' + dec_src[3])/3600.0)
        flux = str(sline[4])
        fluxes.append(float(flux))
        ra_dec = np.array([[ra_deg, dec_deg]])
        try:
            x.append(w.wcs_world2pix(ra_dec, 1)[0][0])
            y.append(w.wcs_world2pix(ra_dec, 1)[0][1])
        except AttributeError:
            x.append(w.wcs_sky2pix(ra_dec, 1)[0][0])
            y.append(w.wcs_sky2pix(ra_dec, 1)[0][1])

    minx = np.min(x)
    miny = np.min(y)
    x += abs(minx)
    y += abs(miny)
    return x, y, np.array(fluxes), arcsec_per_pix


def dist2(x1,y1,x2,y2, scale=1.0) :
    return ((x1-x2)**2 + (y1-y2)**2)/scale**2


def guess_regular_grid(xnodes, ynodes, pixelsize=None) :
    """
    Return a regular grid guessed on an irregular one (Voronoi)
    xnodes, ynodes: arrays of Voronoi bins

    Return: xunb, yunb = regular grid for x and y (unbinned)
    """
    ## First deriving a pixel size
    xn_rav, yn_rav = xnodes.ravel(), ynodes.ravel()
    if pixelsize is None :
        pixelsize = derive_pixelsize(xnodes, ynodes)
    minxn = int(np.min(xn_rav) / pixelsize) * pixelsize
    minyn = int(np.min(yn_rav) / pixelsize) * pixelsize
    xunb, yunb = np.meshgrid(np.arange(minxn, np.max(xn_rav)+pixelsize, pixelsize),
                           np.arange(minyn, np.max(yn_rav)+pixelsize, pixelsize))

    return xunb, yunb


def derive_unbinned_field(xnodes, ynodes, data, xunb=None, yunb=None) :
    """
       Provide an array of the same shape as the input xunb, and yunb
       with the values derived from the Voronoi binned data

       xnodes, ynodes: 2 arrays providing the nodes from the binning
       data : values for each node
       xunb, yunb: x and y coordinates of the unbinned data
                 if not provided (default) they will be guessed from the nodes

       Return: xunb, yunb, and unbinned_data arrays with the same shape as xunb,
    """
    if xunb is None :
        xunb, yunb = guess_regular_grid(xnodes, ynodes)

    x_rav, y_rav = xunb.ravel(), yunb.ravel()
    xnodes_rav, ynodes_rav = xnodes.ravel(), ynodes.ravel()
    data_rav = data.ravel()
    unbinned_data = np.zeros_like(x_rav)
    for i in xrange(len(x_rav)) :
        indclosestBin = argmin(dist2(x_rav[i], y_rav[i], xnodes_rav, ynodes_rav))
        unbinned_data[i] = data_rav[indclosestBin]

    return xunb, yunb, unbinned_data.reshape(xunb.shape)


listmethods = ["voronoi", "quadtree"]
class bin2D:
    """
    Class for Voronoi binning of a set of x and y coordinates
    using given data and potential associated noise
    """
    def __init__(self, xin, yin, data, target_flux=1.0, pixelsize=None,
        method="Voronoi", cvt=1, wvt=0):
        self.xin = xin.ravel()
        self.yin = yin.ravel()
        self.data = data.ravel()

        ## Sort all pixels by their distance to the maximum flux
        if pixelsize is None :
            self.pixelsize = derive_pixelsize(self.xin, self.yin, verbose=1)
        else :
            self.pixelsize = pixelsize

        self.target_flux = target_flux

        ## Binning method and options
        self.method = str.lower(method)
        self.cvt = cvt
        self.wvt = wvt
        self.scale = 1.0
        self._check_input()

    def _check_input(self):
        """
        Check consistency of input data
        """
        # Basic checks of the data
        # First about the dimensions of the datasets
        self.npix = len(self.xin)

    ### Accrete the bins ============================================
    def bin2d_accretion(self, verbose=0):
        """ Accrete the bins according to their flux
        """
        ## Status holds the bin number when assigned
        self.status = np.zeros(self.npix, dtype=Nint)
        ## Good is 1 if the bin was accepted
        self.good = np.zeros(self.xin.size, dtype=Nint)

        ## Start with the max flux in the data
        currentBin = [argmax(self.data)]

        for ind in range(1,self.npix+1):  ## Running over the index of the Voronoi BIN
            ## Only one pixel at this stage
            current_flux = sum(self.data[currentBin])
            if verbose :
                print("Bin %d"%(ind)) # TODO: Change to logging

            self.status[currentBin] = ind   # only one pixel at this stage
            ## Barycentric centroid for 1 pixel...
            xbar, ybar = self.xin[currentBin], self.yin[currentBin]

            ## Indices of remaining unbinned data
            unBinned = np.where(self.status == 0)[0]
            ##+++++++++++++++++++++++++++++++++++++++++++++++++
            ## STOP THE WHILE Loop if all pixels are now binned
            while len(unBinned) > 0 :
            ##+++++++++++++++++++++++++++++++++++++++++++++++++

                ## Coordinates of the Remaining unbinned pixels
                xunBin = self.xin[unBinned]
                yunBin = self.yin[unBinned]

                ## Location of current bin
                xcurrent = self.xin[currentBin]
                ycurrent = self.yin[currentBin]

                ## Closest unbinned pixel to the centroid making sure the highest flux is used
                indclosestBar = argmin(dist2(xbar,ybar, xunBin, yunBin))
                xclosest = xunBin[indclosestBar]
                yclosest = yunBin[indclosestBar]

                ## Distance between this closest pixel and the current pixel
                currentSqrtDist= sqrt(min(dist2(xclosest,yclosest, xcurrent,ycurrent)))

                ## Add new pixel
                possibleBin = currentBin + [unBinned[indclosestBar]]

                ## Transfer new flux to current value
                old_flux = current_flux
                current_flux = sum(self.data[possibleBin])

                ## Test if getting better for flux
                if (current_flux > 0.8 * self.target_flux):
                    if (current_flux < 1.2 * self.target_flux):
                        self.good[possibleBin] = 1
                        break
                    elif (abs(current_flux-self.target_flux) > abs(old_flux-self.target_flux)):
                        if (old_flux > 0.8 * self.target_flux):
                            self.good[currentBin] = 1
                        break
                ##++++++++++++++++++++++++++++++++++++++++++

                ## If the new Bin is ok we associate the number of the bin to that one
                self.status[unBinned[indclosestBar]] = ind
                ##   ... and we use that now as the current Bin
                currentBin = possibleBin

                ## And update the values
                ## First the centroid
                xbar, ybar = mean(self.xin[currentBin]), mean(self.yin[currentBin])

                ## New set of unbinned pixels
                unBinned = np.where(self.status == 0)[0]
                ## ----- End of While Loop --------------------------------------------

            ## Unbinned pixels
            unBinned = np.where(self.status == 0)[0]
            ## Break if all pixels are binned
            if len(unBinned) == 0 :
                break
            ## When the while loop is finished for this new BIN
            ## Find the centroid of all Binned pixels
            Binned = np.where(self.status != 0)[0]
            xbar, ybar = mean(self.xin[Binned]), mean(self.yin[Binned])

            ## Closest unbinned pixel to the centroid of all Binned pixels
            xunBin = self.xin[unBinned]
            yunBin = self.yin[unBinned]
            indclosestBar = argmin(dist2(xbar,ybar, xunBin, yunBin))
            ## Take now this closest pixel as the next one to look at
            currentBin = [unBinned[indclosestBar]]

        ## Set to zero all bins that did not reach the target flux
        self.status *= self.good

    ### Compute centroid of bins ======================================
    def bin2d_centroid(self, verbose=0):
        ## Compute the area for each node
        self.Areanode, self.bins = np.histogram(self.status, bins=np.arange(np.max(self.status+0.5))+0.5)
        ## Select the ones which have at least one bin
        self.indgoodbins = np.where(self.Areanode > 0)[0]
        self.Areanode = self.Areanode[self.indgoodbins]
        ngoodbins = self.indgoodbins.size
        ## Reset the xnode, ynode, SNnode, and statusnode (which provides the number for the node)
        self.xnode = np.zeros(ngoodbins, Nfloat)
        self.ynode = np.zeros_like(self.xnode)
        self.flux_node = np.zeros_like(self.xnode)
        self.statusnode = np.zeros_like(self.xnode)
        self.listbins = []
        for i in range(ngoodbins) :
            ## indgoodbins[i] provides the bin of the Areanode, so indgoodbins[i] + 1 is the status number
            self.statusnode[i] = self.indgoodbins[i]+1
            ## List of bins which have statusnode as a status
            listbins = np.where(self.status==self.statusnode[i])[0]
            ## Centroid of the node
            self.xnode[i], self.ynode[i] = mean(self.xin[listbins]), mean(self.yin[listbins])
            self.flux_node[i] = sum(self.data[listbins])
            self.listbins.append(listbins)

    ### Compute WEIGHTED centroid of bins ======================================
    def bin2d_weighted_centroid(self, weight=None, verbose=0):
        if weight is not None : self.weight = weight

        self.Areanode, self.bins = np.histogram(self.status, bins=np.arange(np.max(self.status+0.5))+0.5)
        ## Select the ones which have at least one bin
        self.indgoodbins = np.where(self.Areanode > 0)[0]
        self.Areanode = self.Areanode[self.indgoodbins]
        ngoodbins = self.indgoodbins.size
        ## Reset the xnode, ynode, SNnode, and statusnode (which provides the number for the node)
        self.xnode = np.zeros(ngoodbins, Nfloat)
        self.ynode = np.zeros_like(self.xnode)
        self.flux_node = np.zeros_like(self.xnode)
        self.statusnode = np.zeros_like(self.xnode)
        self.listbins = []
        for i in range(ngoodbins) :
            ## indgoodbins[i] provides the bin of the Areanode, so indgoodbins[i] + 1 is the status number
            self.statusnode[i] = self.indgoodbins[i]+1
            ## List of bins which have statusnode as a status
            listbins = np.where(self.status==self.statusnode[i])[0]
            ## Weighted centroid of the node
            self.xnode[i], self.ynode[i] = np.average(self.xin[listbins], weights=self.weight[listbins]), np.average(self.yin[listbins], weights=self.weight[listbins])
            self.flux_node[i] = sum(self.data[listbins])
            self.listbins.append(listbins)

    ### Assign bins ============================================
    def bin2d_assign_bins(self, sel_pixels=None, scale=None, verbose=0) :
        """
        Assign the bins when the nodes are derived With Scaling factor
        """
        if scale is not None: self.scale = scale
        if sel_pixels is None : sel_pixels = range(self.xin.size)
        for i in sel_pixels :
            minind = argmin(dist2(self.xin[i], self.yin[i], self.xnode, self.ynode, scale=self.scale))
            self.status[i] = self.statusnode[minind]
            if verbose :
                print("Pixel ",  self.status[i], self.xin[i], self.yin[i], self.xnode[minind], self.ynode[minind]) # TODO: Change to logging

        ## reDerive the centroid
        self.bin2d_centroid()

    ### Do  CV tesselation ============================================
    def bin2d_cvt_equal_mass(self, wvt=None, verbose=1) :
        """
        Produce a CV Tesselation

        wvt: default is None (will use preset value, see self.wvt)
        """

        ## Reset the status and statusnode for all nodes
        self.status = np.zeros(self.npix, dtype=Nint)
        self.statusnode = np.arange(self.xnode.size) + 1

        if wvt is not None : self.wvt = wvt
        if self.wvt : self.weight = np.ones_like(self.data)
        else : self.weight = self.data**4

        self.scale = 1.0

        self.niter = 0
        ## WHILE LOOP: stop when the nodes do not move anymore ============
        Oldxnode, Oldynode = copy.copy(self.xnode[-1]), copy.copy(self.ynode[-1])
        while (not np.array_equiv(self.xnode, Oldxnode)) | (not np.array_equiv(self.ynode, Oldynode)):
            Oldxnode, Oldynode = copy.copy(self.xnode), copy.copy(self.ynode)
            ## Assign the closest centroid to each bin
            self.bin2d_assign_bins()

            ## New nodes weighted centroids
            self.bin2d_weighted_centroid()

            ## Eq. (4) of Diehl & Statler (2006)
            if self.wvt : self.scale = sqrt(self.Areanode/self.flux_node)
            self.niter += 1

    ### Do Voronoi binning ============================================
    def bin_voronoi(self, wvt=None, cvt=None, verbose=0) :
        """ Actually do the Voronoi binning

        wvt: default is None (will use preset value, see self.wvt)
        cvt: default is None (will use preset value, see self.cvt)
        """
        if cvt is not None : self.cvt = cvt
        if wvt is not None : self.wvt = wvt

        self.bin2d_accretion(verbose=verbose)
        self.bin2d_centroid()
        ## Get the bad pixels, not assigned and assign them
        badpixels = np.where(self.status == 0)[0]
        self.bin2d_assign_bins(badpixels)
        if self.cvt :
            self.bin2d_cvt_equal_mass()
        else : self.scale = 1.0
        ## Final nodes weighted centroids after assigning to the final nodes
        self.bin2d_assign_bins()
        if self.wvt : self.weight = np.ones_like(self.data)
        else : self.weight = self.data
        self.bin2d_weighted_centroid()


    def show_voronoibin(self, datain=None, shownode=1, mycmap=None) :
        """
        Display the voronoi bins on a map

        datain: if None (Default), will use random colors to display the bins
                if provided, will display that with a jet (or specified mycmap) cmap
                   (should be either the length of the voronoi nodes array or the size of the initial pixels)
        shownode: default is 1 -> show the voronoi nodes, otherwise ignore (0)
        mycmap: in case datain is provide, will use that cmpa to display the bins
        """
        from distutils import version
        try:
            import matplotlib
        except ImportError:
            raise Exception("matplotlib 0.99.0 or later is required for this routine")

        if version.LooseVersion(matplotlib.__version__) < version.LooseVersion('0.99.0'):
            raise Exception("matplotlib 0.99.0 or later is required for this routine")

        from matplotlib.collections import PatchCollection
        import matplotlib.patches as mpatches
        import matplotlib.pyplot as plt
        fig = plt.figure(1,figsize=(7,7))
        plt.clf()
        ax = plt.gca()
        patches = []
        binsize = self.pixelsize
        for i in range(len(self.xin)) :
            patches.append(mpatches.Rectangle((self.xin[i],self.yin[i]), binsize*10, binsize*10))

        if datain is None :
            dataout = self.status
            mycmap = 'prism'
        else :
            if len(datain) == self.xnode.size :
                dataout = np.zeros(self.xin.size, Nfloat)
                for i in range(self.xnode.size) :
                    listbins = self.listbins[i]
                    dataout[listbins] = [datain[i]]*len(listbins)
            elif len(datain) == self.xin.size :
                dataout = datain
            if mycmap is None : mycmap = 'jet'

        colors = dataout * 100.0 / max(dataout)
        collection = PatchCollection(patches, cmap=mycmap)
        collection.set_array(np.array(colors))
        ax.add_collection(collection)
        if shownode :
            plt.scatter(self.xnode,self.ynode, marker='o', edgecolors='k', facecolors='none', s=100)
            for i in range(self.xnode.size):
                ax.annotate(i, (self.xnode[i], self.ynode[i]))
                listbins = self.listbins[i]

        plt.axis('image')
        plt.xlabel("X axis")
        plt.ylabel("Y axis")
        plt.title("Voronoi Map")

    def bin_data(self, datain=None, noisein=None) :
        """
        Return a Voronoi adaptive binning of your data.

        datain: if provided, will be used as data input
                if not provided (None = default), will use self.data
        noisein: if provided, will be used as noise input
                if not provided (None = default), will use self.noise

        Output = xnode, ynode, bindata, S/N
        """

        if datain is None: datain = copy.copy(self.data)

        dataout = np.zeros(self.xnode.size, Nfloat)
        xout = np.zeros_like(dataout)
        yout = np.zeros_like(dataout)
        flux_out = np.zeros_like(dataout)
        for i in range(self.xnode.size) :
            listbins = self.listbins[i]
            xout[i] = np.average(self.xin.ravel()[listbins], weights=datain[listbins])
            yout[i] = np.average(self.yin.ravel()[listbins], weights=datain[listbins])
            dataout[i] = mean(datain[listbins])
            flux_out[i] = sum(datain[listbins])

        return xout, yout, dataout, flux_out


def derive_pixelsize(x, y, verbose=0) :
    """ Find the pixelsize by looking at the minimum distance between
        pairs of x,y
        x: xaxis coordinates
        y: yaxis coordinates
        Return: pixelsize
    """
    pixelsize = 1.e30
    for i in range(len(x)-1) :
        mindist = np.min(dist2(x[i], y[i], x[i+1:], y[i+1:]))
        pixelsize = np.minimum(mindist, pixelsize)
    pixelsize = np.sqrt(pixelsize)
    return pixelsize


def bins2Patches(bin_obj, root='Patch', pad_index=False):
    """Returns a patch column based on binning"""
    patchNames = []
    for src_idx in range(len(bin_obj.data)):
        for i in range(bin_obj.xnode.size):
            listbins = bin_obj.listbins[i]
            if src_idx in listbins:
                if pad_index:
                    patchNames.append('{0}_{1}'.format(root,
                        str(i).zfill(int(np.ceil(np.log10(bin_obj.xnode.size))))))
                else:
                    patchNames.append('{0}_{1}'.format(root, i))
    return np.array(patchNames)




# -*- coding: utf-8 -*-
#
# Defines threshold functions used by the group operation
#
# Copyright (C) 2013 -  David Rafferty
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


def getPatchNamesByThreshold(LSM, fwhmArcsec, threshold=0.1, root='threshold',
    pad_index=False):
    """
    Projects sky model to image plane, convolves with Gaussian, and finds islands
    of emission
    """
    import numpy as np

    LSM.ungroup()

    # Generate image grid with 1 pix = FWHM / 4
    x, y, midRA, midDec = LSM._getXY(crdelt=fwhmArcsec/4.0/3600.0)
    sizeX = int(1.2 * (max(x) - min(x)))
    sizeY = int(1.2 * (max(y) - min(y)))
    image = np.zeros((sizeX, sizeY))
    xint = np.array(x, dtype=int)
    xint += -1 * min(xint) + 1
    yint = np.array(y, dtype=int)
    yint += -1 * min(yint) + 1

    # Set pixels with sources to one
    image[xint, yint] = 1.0

    # Convolve with Gaussian of FWHM = 4 pixels
    image = blur_image(image, 4.0/2.35482)

    mask = image / threshold >= 1.0
    patchCol = getPatchNamesFromMask(mask, xint, yint, root=root, pad_index=pad_index)

    return patchCol


def blur_image(im, n, ny=None):
    """
    Blurs the image by convolving with a gaussian kernel of typical
    size n. The optional keyword argument ny allows for a different
    size in the y direction.
    """
    from scipy.ndimage import gaussian_filter

    sx = n
    if ny is not None:
        sy = ny
    else:
        sy = n
    improc = gaussian_filter(im, [sy, sx])

    return improc


def getPatchNamesFromMask(mask, x, y, root='mask', pad_index=False):
    """
    Returns an array of patch names for each (x, y) pair
    """
    import scipy.ndimage as nd
    import numpy as np

    act_pixels = mask
    rank = len(act_pixels.shape)
    connectivity = nd.generate_binary_structure(rank, rank)
    mask_labels, count = nd.label(act_pixels, connectivity)

    patchNums = []
    patchNames = []
    for xs, ys in zip(x, y):
        try:
            patchNums.append(mask_labels[xs, ys])
        except:
            patchNums.append(0)

    # Check if there is a patch with id = 0. If so, this means there were
    # some Gaussians that fell outside of the regions in the patch
    # mask file.
    n = 0
    for p in patchNums:
        if p != 0:
            if pad_index:
                patchNames.append('{0}_patch_'.format(root)+
                    str(p).zfill(int(np.ceil(np.log10(len(set(patchNums))+1)))))
            else:
                patchNames.append('{0}_patch_'.format(root)+str(p))
        else:
            patchNames.append('patch_'+str(n))
            n += 1

    return np.array(patchNames)



#!/usr/bin/env python3

"""Compute the absolute Gaussian position angle for a skymodel file"""

from astropy.table import Table
from argparse import ArgumentParser
from astropy.coordinates import Angle
import astropy.units as u
from lsmtool import tableio  # Registers makesourcedb reader and writer with astropy.tables
from numpy import sin, cos, arctan2
import logging


logger = logging.getLogger(__name__)

def compute_absolute_orientation(
    relative_orientation, ra_source, dec_source, ra_center, dec_center
):
    """Compute the absolute position angle from a relative one, where the relative one was created from an image with center ra_center, dec_center.

    Args:
        relative_orientation (astropy Quantity): position angle with respect to Norh in the image
        ra_center (astropy Quantity): right ascension of the center of the image from which the relative position angle was deduced
        dec_center (astropy Quantity): declination of the center of the image from which the relative position angle was deduced
    """
    ra_diff = ra_source - ra_center
    dl_ddec = -sin(dec_source) * sin(ra_diff)
    dm_ddec = cos(dec_source) * cos(dec_center) + sin(dec_source) * sin(dec_center) * cos(ra_diff)
    angle_towards_ncp = arctan2(dl_ddec, dm_ddec)
    return relative_orientation - angle_towards_ncp


def add_absolute_orientation(skymodel_filename_in, skymodel_filename_out, ra_center, dec_center):
    """Add absolute position for all gaussian sources"""
    skymodel_table = Table.read(skymodel_filename_in, format="makesourcedb")
    orientation_is_absolute_column_present = "OrientationIsAbsolute" in skymodel_table.columns
    if "Orientation" not in skymodel_table.columns:
        raise RuntimeError("No Orientation column present in " + skymodel_filename_in)
    if not orientation_is_absolute_column_present:
        skymodel_table["OrientationIsAbsolute"] = False
    for row in skymodel_table:
        if row["Type"] == "GAUSSIAN":
            if row["OrientationIsAbsolute"]:
                print("Orientation is already absolute, skipping")
            else:
                absolute_orientation = compute_absolute_orientation(
                    row["Orientation"] * u.deg,
                    row["Ra"] * u.deg,
                    row["Dec"] * u.deg,
                    ra_center,
                    dec_center,
                )
            row["Orientation"] = absolute_orientation.to(u.deg).value
            row["OrientationIsAbsolute"] = True
    skymodel_table.write(skymodel_filename_out, format="makesourcedb")


if __name__ == "__main__":
    parser = ArgumentParser(description="Add absolute Gaussian position angle to a skymodel file")
    parser.add_argument("skymodelfile", help="Skymodel file")
    parser.add_argument(
        "ra_center", help="Right ascension of center, e.g. '32.2deg' or '3h35m23.2'", type=str
    )
    parser.add_argument(
        "dec_center", help="Declination of center, e.g. '12.5deg' or '5d35m17.8'", type=str
    )
    parser.add_argument(
        "-o", "--output", help="Output filename (default: derived from input filename)", type=str
    )
    args = parser.parse_args()

    skymodel_filename_out = args.output
    if skymodel_filename_out is None:
        if args.skymodelfile.endswith(".skymodel") or args.skymodelfile.endswith(".txt"):
            filename_parts = args.skymodelfile.rsplit(".")
            skymodel_filename_out = filename_parts[0] + "_absolute_orientation." + filename_parts[1]
        else:
            skymodel_filename_out = filename + ".absolute_orientation"
    if args.ra_center.count(':') == 2:  # Handle casacore format in e.g. msoverview output
        args.ra_center = args.ra_center.replace(":", "h", 1).replace(":", "m", 1)
    if args.dec_center.count(".") == 3:
        args.dec_center = args.dec_center.replace(".", "d", 1).replace(".", "m", 1)
    ra_center = Angle(args.ra_center)
    dec_center = Angle(args.dec_center)
    add_absolute_orientation(args.skymodelfile, skymodel_filename_out, ra_center, dec_center)
    logger.info("Saved corrected skymodel to " + skymodel_filename_out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation compares two sky models
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging

log = logging.getLogger('LSMTool.COMPARE')
log.debug('Loading COMPARE module.')


def run(step, parset, LSM):

    outDir = parset.getString('.'.join(["LSMTool.Steps", step, "OutDir"]), '')
    skyModel2 = parset.getString('.'.join(["LSMTool.Steps", step, "SkyModel2"]), '')
    radius = parset.getString('.'.join(["LSMTool.Steps", step, "Radius"]), '10 arcsec')
    labelBy = parset.getString('.'.join(["LSMTool.Steps", step, "LabelBy"]), '')
    excludeMultiple = parset.getBool('.'.join(["LSMTool.Steps", step, "ExcludeMultiple"]), True)
    ignoreSpec = parset.getString('.'.join(["LSMTool.Steps", step, "IgnoreSpec"]), '')

    if outDir == '':
        outDir = '.'
    if labelBy == '':
        labelBy = None
    if ignoreSpec == '':
        ignoreSpec = None
    else:
        ignoreSpec = float(ignoreSpec)

    try:
        compare(LSM, skyModel2, radius=radius, outDir=outDir, labelBy=labelBy,
                excludeMultiple=excludeMultiple, ignoreSpec=ignoreSpec)
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    return result


def compare(LSM1, LSM2, radius='10 arcsec', outDir='.', labelBy=None,
            ignoreSpec=None, excludeMultiple=True, excludeByFlux=True,
            name1=None, name2=None, format='pdf', make_plots=True):
    """
    Compare two sky models

    Comparison plots and a text file with statistics are written out to the
    an output directory. Plots are made for:
        - flux ratio vs. radius from sky model center
        - flux ratio vs. sky position
        - flux ratio vs flux
        - position offsets
    The following statistics are saved to 'stats.txt' in the output directory:
        - mean and standard deviation of flux ratio
        - mean and standard deviation of RA offsets (in degrees)
        - mean and standard deviation of Dec offsets (in degrees)

    Parameters
    ----------
    LSM1 : SkyModel object
        Parent sky model
    LSM2 : SkyModel object
        Sky model to compare to the parent sky model
    radius : float or str, optional
        Radius in degrees (if float) or 'value unit' (if str; e.g., '30 arcsec')
        for matching
    outDir : str, optional
        Plots are saved to this directory
    labelBy : str, optional
        One of 'source' or 'patch': label points using source names ('source') or
        patch names ('patch')
    ignoreSpec : float, optional
        Ignore sources with this spectral index
    excludeMultiple : bool, optional
        If True, sources with multiple matches are excluded. If False, the
        nearest of the multiple matches will be used for comparison
    excludeByFlux : bool, optional
        If True, matches whose predicted fluxes differ from the parent model
        fluxes by 25% are excluded from the positional offset plot.
    name1 : str, optional
        Name to use in the plots for LSM1. If None, 'Model 1' is used.
    name2 : str, optional
        Name to use in the plots for LSM2. If None, 'Model 2' is used.
    format : str, optional
        Format of plot files.
    make_plots : bool, optional
        If True, the plots described above are made.

    Examples
    --------
    Compare two sky models and save plots and stats to 'comparison_results/'::

        >>> LSM1 = lsmtool.load('sky1.model')
        >>> LSM2 = lsmtool.load('sky2.model')
        >>> compare(LSM1, LSM2, outDir='comparison_results/')

    Compare a LOFAR sky model to a global sky model made from VLSS+TGSS+NVSS (where
    refRA and refDec are the approximate center of the LOFAR sky model coverage)::

        >>> LSM1 = lsmtool.load('lofar_sky.model')
        >>> LSM2 = lsmtool.load('GSM', VOPosition=[refRA, refDec], VORadius='5 deg')
        >>> compare(LSM1, LSM2, radius='30 arcsec', excludeMultiple=True,
            outDir='comparison_results/', name1='LOFAR', name2='GSM', format='png')

    """
    from ..operations_lib import matchSky, radec2xy
    from ..skymodel import SkyModel
    import numpy as np
    import os

    if type(LSM2) is str:
        LSM2 = SkyModel(LSM2)

    if len(LSM1) == 0:
        log.info('Parent sky model is empty. No comparison possible.')
        return
    if len(LSM2) == 0:
        log.info('Secondary sky model is empty. No comparison possible.')
        return

    byPatch = False
    if (LSM1.hasPatches and not LSM2.hasPatches):
        LSM2.group('every')
    if (LSM2.hasPatches and not LSM1.hasPatches):
        LSM1.group('every')
    if (LSM2.hasPatches and LSM1.hasPatches):
        byPatch = True

    # Cross match the tables
    if excludeMultiple:
        nearestOnly = False
    else:
        nearestOnly = True
    matches11, matches21 = matchSky(LSM1, LSM2, radius=radius, byPatch=byPatch,
                                    nearestOnly=nearestOnly)
    matches12, matches22 = matchSky(LSM2, LSM1, radius=radius, byPatch=byPatch,
                                    nearestOnly=nearestOnly)
    if len(matches11) == 0:
        log.info('No matches found.')
        return

    # Get reference frequencies
    if byPatch:
        aggregate = 'wmean'
    else:
        aggregate = None
    if 'ReferenceFrequency' in LSM1.getColNames():
        refFreq1 = LSM1.getColValues('ReferenceFrequency', aggregate=aggregate)
    else:
        refFreq1 = np.array([LSM1.table.meta['ReferenceFrequency']]*len(LSM1))
    if 'ReferenceFrequency' in LSM2.getColNames():
        refFreq2 = LSM2.getColValues('ReferenceFrequency', aggregate=aggregate)
    else:
        refFreq2 = np.array([LSM2.table.meta['ReferenceFrequency']]*len(LSM2))

    # For simplicity, consider only logarithmic indices and only the first term.
    # For sources with no or non-logarithmic indices, adopt the typical value
    # of -0.8
    try:
        # Get the spectral indices and keep first term only
        alphas2 = LSM2.getColValues('SpectralIndex', aggregate=aggregate)
        if len(alphas2.shape) > 1:
            if alphas2.shape[1] == 1:
                # Remove extra axis
                alphas2 = alphas2.squeeze(axis=1)
            else:
                # Take only the first term
                alphas2 = alphas2[:, 0]
    except (IndexError, ValueError):
        # No indices in table, so use typical value for alpha of -0.8 for all
        # sources
        alphas2 = np.array([-0.8]*len(LSM2))
    if ('LogarithmicSI' in LSM2.getColNames() and
            np.any(LSM2.getColValues('LogarithmicSI') == 'false')):
        # One or more sources have non-logarithmic indices, so adopt typical
        # value for alpha of -0.8 for those sources
        logsi = LSM2.getColValues('LogarithmicSI')
        if byPatch:
            patch_names = LSM2.getColValues('Patch')
            logsi_false_patch_names = set(patch_names[np.where(logsi == 'false')])
            for patch in logsi_false_patch_names:
                alphas2[patch_names.tolist().index(patch)] = -0.8
        else:
            alphas2[np.where(logsi == 'flase')] = -0.8

    # Select sources that match up with only a single source and filter by spectral
    # index if desired
    filter = []
    for i in range(len(matches11)):
        nMatches = len(np.where(matches21 == matches21[i])[0])
        if nMatches == 1:
            # This source has a single match
            if ignoreSpec is not None:
                spec = alphas2[matches21[i]]
                if spec != ignoreSpec:
                    filter.append(i)
            else:
                filter.append(i)
    good1 = set(matches11[filter])
    filter = []
    for i in range(len(matches12)):
        nMatches = len(np.where(matches22 == matches22[i])[0])
        if nMatches == 1:
            # This source has a single match
            if ignoreSpec is not None:
                spec = alphas2[matches12[i]]
                if spec != ignoreSpec:
                    filter.append(i)
            else:
                filter.append(i)
    good2 = set(matches22[filter])
    good = good1.intersection(good2)
    matches1 = []
    matches2 = []
    for i in range(len(matches11)):
        if matches11[i] in good:
            matches1.append(matches11[i])
            matches2.append(matches21[i])
    if len(matches1) == 0:
        log.info('No suitable sources found for comparison.')
        return

    # Apply the filters
    if byPatch:
        fluxes1 = LSM1.getColValues('I', aggregate='sum')[matches1]
        fluxes2 = LSM2.getColValues('I', aggregate='sum')[matches2]
        RA, Dec = LSM1.getPatchPositions(asArray=True)
        RA = RA[matches1]
        Dec = Dec[matches1]
        RA2, Dec2 = LSM2.getPatchPositions(asArray=True)
        RA2 = RA2[matches2]
        Dec2 = Dec2[matches2]
    else:
        fluxes1 = LSM1.getColValues('I', aggregate=aggregate)[matches1]
        fluxes2 = LSM2.getColValues('I', aggregate=aggregate)[matches2]
        RA = LSM1.getColValues('Ra')[matches1]
        Dec = LSM1.getColValues('Dec')[matches1]
        RA2 = LSM2.getColValues('Ra')[matches2]
        Dec2 = LSM2.getColValues('Dec')[matches2]

    # Calculate predicted LSM2 fluxes at frequencies of LSM1
    predFlux = fluxes2
    predFlux *= 10.0**(alphas2[matches2] * np.log10(refFreq1[matches1] / refFreq2[matches2]))

    # Find reference RA and Dec for center of LSM1
    x, y, refRA, refDec = LSM1._getXY()
    if byPatch:
        x, y = radec2xy(RA, Dec, refRA, refDec)
    else:
        x = x[matches1]
        y = y[matches1]
    refx, refy = radec2xy(RA2, Dec2, refRA, refDec)

    if labelBy is not None:
        if labelBy.lower() == 'source':
            labels = LSM1.getColValues('name')[matches1]
        elif labelBy.lower() == 'patch':
            if LSM1.hasPatches:
                labels = LSM1.getPatchNames()[matches1]
            else:
                labels = LSM1.getColValues('name')[matches1]
        else:
            raise ValueError("The lableBy parameter must be one of 'source' or "
                             "'patch'.")
    else:
        labels = None

    # Make plots
    if outDir is None:
        outDir = '.'
    if outDir[-1] != '/':
        outDir += '/'
    if not os.path.exists(outDir):
        os.makedirs(outDir)
    if make_plots:
        plotFluxRatiosDist(predFlux, fluxes1, RA, Dec, refRA, refDec, labels, outDir, name1, name2, format)
        plotFluxRatioSky(predFlux, fluxes1, x, y, RA, Dec, refRA, refDec, labels, outDir, name1, name2, format)
        plotFluxRatiosFlux(predFlux, fluxes1, labels, outDir, name1, name2, format)
        retstatus = plotOffsets(RA, Dec, RA2, Dec2, x, y, refx, refy, labels, outDir, predFlux,
                                fluxes1, excludeByFlux, name1, name2, format)
        if retstatus == 1:
            log.warn('No matches found within +/- 25% of predicted flux density. Skipping offset plot.')
    argInfo = 'Used radius = {0}, ignoreSpec = {1}, and excludeMultiple = {2}'.format(radius, ignoreSpec,
                                                                                      excludeMultiple)
    stats = findStats(predFlux, fluxes1, RA, Dec, RA2, Dec2, outDir, argInfo,
                      LSM1._info(), LSM2._info(), name1, name2)

    return stats


def plotFluxRatiosDist(predFlux, measFlux, RA, Dec, refRA, refDec, labels, outDir,
                       name1, name2, format, clip=True):
    """
    Makes plot of measured-to-predicted flux ratio vs. distance from center
    """
    import numpy as np
    from ..operations_lib import calculateSeparation
    try:
        from astropy.stats.funcs import sigma_clip
    except ImportError:
        from astropy.stats import sigma_clip
    try:
        import matplotlib
        if matplotlib.get_backend() != 'Agg':
            matplotlib.use("Agg")
        import matplotlib.pyplot as plt
    except Exception as e:
        raise ImportError('PyPlot could not be imported. Plotting is not '
                          'available: {0}'.format(e))

    if name1 is None:
        name1 = 'Model 1'
    if name2 is None:
        name2 = 'Model 2'

    ratio = measFlux / predFlux
    separation = np.zeros(len(measFlux))
    for i in range(len(measFlux)):
        separation[i] = calculateSeparation(RA[i], Dec[i], refRA, refDec).value

    fig = plt.figure(figsize=(7.0, 5.0))
    ax1 = plt.subplot(1, 1, 1)
    ax1.plot(separation, ratio, 'o')
    plt.title('Flux Density Ratios ({0} / {1})'.format(name1, name2))
    plt.ylabel('Flux density ratio')
    plt.xlabel('Distance from center (deg)')

    # Calculate mean ratio and std dev.
    if clip:
        mean_ratio = np.mean(sigma_clip(ratio))
        std = np.std(sigma_clip(ratio))
    else:
        mean_ratio = np.mean(ratio)
        std = np.std(ratio)
    ax1.set_ylim(0, 2.0*mean_ratio)
    xmin, xmax, ymin, ymax = plt.axis()
    ax1.plot([0.0, xmax], [mean_ratio, mean_ratio], '--g')
    ax1.plot([0.0, xmax], [mean_ratio+std, mean_ratio+std], '-.g')
    ax1.plot([0.0, xmax], [mean_ratio-std, mean_ratio-std], '-.g')

    if labels is not None:
        xls = separation
        yls = ratio
        for label, xl, yl in zip(labels, xls, yls):
            plt.annotate(label, xy=(xl, yl), xytext=(-2, 2), textcoords='offset points',
                         ha='right', va='bottom')

    plt.savefig(outDir+'flux_ratio_vs_distance.{}'.format(format), format=format)
    plt.close('all')


def plotFluxRatiosFlux(predFlux, measFlux, labels, outDir, name1, name2, format, clip=True):
    """
    Makes plot of measured-to-predicted flux ratio vs. flux
    """
    import numpy as np
    try:
        from astropy.stats.funcs import sigma_clip
    except ImportError:
        from astropy.stats import sigma_clip
    try:
        import matplotlib
        if matplotlib.get_backend() != 'Agg':
            matplotlib.use("Agg")
        import matplotlib.pyplot as plt
    except Exception as e:
        raise ImportError('PyPlot could not be imported. Plotting is not '
                          'available: {0}'.format(e))

    if name1 is None:
        name1 = 'Model 1'
    if name2 is None:
        name2 = 'Model 2'

    ratio = measFlux / predFlux

    fig = plt.figure(figsize=(7.0, 5.0))
    ax1 = plt.subplot(1, 1, 1)
    ax1.plot(measFlux, ratio, 'o')
    ax1.set_xscale('log')
    plt.title('Flux Density Ratios ({0} / {1})'.format(name1, name2))
    plt.ylabel('Flux density ratio')
    plt.xlabel('{0} flux density (Jy)'.format(name1))

    # Calculate mean ratio and std dev.
    if clip:
        mean_ratio = np.mean(sigma_clip(ratio))
        std = np.std(sigma_clip(ratio))
    else:
        mean_ratio = np.mean(ratio)
        std = np.std(ratio)
    ax1.set_ylim(0, 2.0*mean_ratio)
    xmin, xmax, ymin, ymax = plt.axis()
    ax1.plot([0.0, xmax], [mean_ratio, mean_ratio], '--g')
    ax1.plot([0.0, xmax], [mean_ratio+std, mean_ratio+std], '-.g')
    ax1.plot([0.0, xmax], [mean_ratio-std, mean_ratio-std], '-.g')

    if labels is not None:
        xls = measFlux
        yls = ratio
        for label, xl, yl in zip(labels, xls, yls):
            plt.annotate(label, xy=(xl, yl), xytext=(-2, 2), textcoords='offset points',
                         ha='right', va='bottom')

    plt.savefig(outDir+'flux_ratio_vs_flux.{}'.format(format), format=format)
    plt.close('all')


def plotFluxRatioSky(predFlux, measFlux, x, y, RA, Dec, midRA, midDec, labels,
                     outDir, name1, name2, format):
    """
    Makes sky plot of measured-to-predicted flux ratio
    """
    import numpy as np
    from ..operations_lib import makeWCS
    try:
        from astropy.stats.funcs import sigma_clip
    except ImportError:
        from astropy.stats import sigma_clip
    try:
        import matplotlib
        if matplotlib.get_backend() != 'Agg':
            matplotlib.use("Agg")
        import matplotlib.pyplot as plt
        import matplotlib.colors as colors
    except Exception as e:
        raise ImportError('PyPlot could not be imported. Plotting is not '
                          'available: {0}'.format(e))
    from astropy.visualization.wcsaxes import WCSAxes

    if name1 is None:
        name1 = 'Model 1'
    if name2 is None:
        name2 = 'Model 2'

    ratio = measFlux / predFlux

    fig = plt.figure(figsize=(7.0, 5.0))
    wcs = makeWCS(midRA, midDec)
    ax1 = WCSAxes(fig, [0.12, 0.12, 0.8, 0.8], wcs=wcs)
    fig.add_axes(ax1)
    plt.title('Flux Density Ratios ({0} / {1})'.format(name1, name2))

    # Set symbol color by ratio
    vmin = np.min(ratio) - 0.1
    vmax = np.max(ratio) + 0.1
    sm = plt.cm.ScalarMappable(cmap=plt.cm.jet, norm=colors.Normalize(vmin=vmin, vmax=vmax))
    sm.set_array(ratio)
    sm._A = []
    c = []
    for r in ratio:
        c.append(sm.to_rgba(r))

    ax1.set_xlim(np.min(x)-20, np.max(x)+20)
    ax1.set_ylim(np.min(y)-20, np.max(y)+20)
    plot = plt.scatter(x, y, c=c)
    cbar = plt.colorbar(sm)

    # Set axis labels, etc.
    RAAxis = ax1.coords['ra']
    DecAxis = ax1.coords['dec']
    RAAxis.set_axislabel('RA')
    DecAxis.set_axislabel('Dec')
    ax1.coords.grid(color='black', alpha=0.5, linestyle='solid')

    if labels is not None:
        xls = x
        yls = y
        for label, xl, yl in zip(labels, xls, yls):
            plt.annotate(label, xy=(xl, yl), xytext=(-2, 2), textcoords='offset points',
                         ha='right', va='bottom')

    plt.savefig(outDir+'flux_ratio_sky.{}'.format(format), format=format)
    plt.close('all')


def plotOffsets(RA, Dec, refRA, refDec, x, y, refx, refy, labels, outDir,
                predFlux, measFlux, excludeByFlux, name1, name2, format,
                plot_imcoords=False):
    """
    Makes plot of measured - predicted RA and DEC and x and y offsets
    """
    from ..operations_lib import calculateSeparation
    import numpy as np
    try:
        from astropy.stats.funcs import sigma_clip
    except ImportError:
        from astropy.stats import sigma_clip
    try:
        import matplotlib
        if matplotlib.get_backend() != 'Agg':
            matplotlib.use("Agg")
        import matplotlib.pyplot as plt
    except Exception as e:
        raise ImportError('PyPlot could not be imported. Plotting is not '
                          'available: {0}'.format(e))

    if excludeByFlux:
        ratio = measFlux / predFlux
        goodInd = np.where((ratio > 0.75) & (ratio < 1.25))[0]
        if len(goodInd) == 0:
            return 1
        RA = RA[goodInd]
        Dec = Dec[goodInd]
        refRA = refRA[goodInd]
        refDec = refDec[goodInd]

    if name1 is None:
        name1 = 'Model 1'
    if name2 is None:
        name2 = 'Model 2'

    RAOffsets = np.zeros(len(RA))
    DecOffsets = np.zeros(len(Dec))
    xOffsets = np.zeros(len(RA))
    yOffsets = np.zeros(len(Dec))
    for i in range(len(RA)):
        if RA[i] >= refRA[i]:
            sign = 1.0
        else:
            sign = -1.0
        RAOffsets[i] = sign * calculateSeparation(RA[i], Dec[i], refRA[i], Dec[i]).value * 3600.0  # arcsec
        xOffsets[i] = x[i] - refx[i]
        if Dec[i] >= refDec[i]:
            sign = 1.0
        else:
            sign = -1.0
        DecOffsets[i] = sign * calculateSeparation(RA[i], Dec[i], RA[i], refDec[i]).value * 3600.0  # arcsec
        yOffsets[i] = y[i] - refy[i]

    fig = plt.figure(figsize=(7.0, 5.0))
    ax1 = plt.subplot(1, 1, 1)
    plt.title('Positional offsets ({0} - {1})'.format(name1, name2))
    ax1.plot(RAOffsets, DecOffsets, 'o')
    xmin, xmax, ymin, ymax = plt.axis()
    ax1.plot([xmin, xmax], [0.0, 0.0], '--g')
    ax1.plot([0.0, 0.0], [ymin, ymax], '--g')
    plt.xlabel('RA Offset (arcsec)')
    plt.ylabel('Dec Offset (arcsec)')

    if labels is not None:
        xls = RAOffsets
        yls = DecOffsets
        for label, xl, yl in zip(labels, xls, yls):
            plt.annotate(label, xy=(xl, yl), xytext=(-2, 2), textcoords='offset points',
                         ha='right', va='bottom')
    plt.savefig(outDir+'positional_offsets_sky.{}'.format(format), format=format)

    if plot_imcoords:
        fig = plt.figure(figsize=(7.0, 5.0))
        ax1 = plt.subplot(1, 1, 1)
        plt.title('Positional offsets ({0} - {1})'.format(name1, name2))
        ax1.plot(xOffsets, yOffsets, 'o')
        xmin, xmax, ymin, ymax = plt.axis()
        ax1.plot([xmin, xmax], [0.0, 0.0], '--g')
        ax1.plot([0.0, 0.0], [ymin, ymax], '--g')
        plt.xlabel('Image-plane X Offset (arb. units)')
        plt.ylabel('Image-plane Y Offset (arb. units)')

        if labels is not None:
            xls = RAOffsets
            yls = DecOffsets
            for label, xl, yl in zip(labels, xls, yls):
                plt.annotate(label, xy=(xl, yl), xytext=(-2, 2), textcoords='offset points',
                             ha='right', va='bottom')
        plt.savefig(outDir+'positional_offsets_im.{}'.format(format), format=format)
    return 0


def findStats(predFlux, measFlux, RA, Dec, refRA, refDec, outDir, info0, info1,
              info2, name1, name2):
    """
    Calculates statistics and saves them to 'stats.txt'
    """
    import numpy as np
    from ..operations_lib import calculateSeparation
    try:
        from astropy.stats.funcs import sigma_clip
    except ImportError:
        from astropy.stats import sigma_clip

    ratio = measFlux / predFlux
    meanRatio = np.mean(ratio)
    stdRatio = np.std(ratio)
    clippedRatio = sigma_clip(ratio)
    meanClippedRatio = np.mean(clippedRatio)
    stdClippedRatio = np.std(clippedRatio)

    RAOffsets = np.zeros(len(RA))
    DecOffsets = np.zeros(len(Dec))
    for i in range(len(RA)):
        if RA[i] >= refRA[i]:
            sign = 1.0
        else:
            sign = -1.0
        RAOffsets[i] = sign * calculateSeparation(RA[i], Dec[i], refRA[i], Dec[i]).value  # deg
        if Dec[i] >= refDec[i]:
            sign = 1.0
        else:
            sign = -1.0
        DecOffsets[i] = sign * calculateSeparation(RA[i], Dec[i], RA[i], refDec[i]).value  # deg

    meanRAOffset = np.mean(RAOffsets)
    stdRAOffset = np.std(RAOffsets)
    clippedRAOffsets = sigma_clip(RAOffsets)
    meanClippedRAOffset = np.mean(clippedRAOffsets)
    stdClippedRAOffset = np.std(clippedRAOffsets)

    meanDecOffset = np.mean(DecOffsets)
    stdDecOffset = np.std(DecOffsets)
    clippedDecOffsets = sigma_clip(DecOffsets)
    meanClippedDecOffset = np.mean(clippedDecOffsets)
    stdClippedDecOffset = np.std(clippedDecOffsets)

    stats = {'meanRatio': meanRatio,
             'stdRatio': stdRatio,
             'meanClippedRatio': meanClippedRatio,
             'stdClippedRatio': stdClippedRatio,
             'meanRAOffsetDeg': meanRAOffset,
             'stdRAOffsetDeg': stdRAOffset,
             'meanClippedRAOffsetDeg': meanClippedRAOffset,
             'stdClippedRAOffsetDeg': stdClippedRAOffset,
             'meanDecOffsetDeg': meanDecOffset,
             'stdDecOffsetDeg': stdDecOffset,
             'meanClippedDecOffsetDeg': meanClippedDecOffset,
             'stdClippedDecOffsetDeg': stdClippedDecOffset}

    outLines = ['Statistics from sky model comparison\n']
    outLines.append('------------------------------------\n\n')

    outLines.append('Sky model 1 ({}):\n'.format(name1))
    outLines.append(info1+'\n\n')
    outLines.append('Sky model 2 ({}):\n'.format(name2))
    outLines.append(info2+'\n\n')

    outLines.append(info0+'\n')
    outLines.append('Number of matches found for comparison: {0}\n\n'.format(len(predFlux)))

    outLines.append('Mean flux density ratio (1 / 2): {0}\n'.format(meanRatio))
    outLines.append('Std. dev. flux density ratio (1 / 2): {0}\n'.format(stdRatio))
    outLines.append('Mean 3-sigma-clipped flux density ratio (1 / 2): {0}\n'.format(meanClippedRatio))
    outLines.append('Std. dev. 3-sigma-clipped flux density ratio (1 / 2): {0}\n\n'.format(stdClippedRatio))

    outLines.append('Mean RA offset (1 - 2): {0} degrees\n'.format(meanRAOffset))
    outLines.append('Std. dev. RA offset (1 - 2): {0} degrees\n'.format(stdRAOffset))
    outLines.append('Mean 3-sigma-clipped RA offset (1 - 2): {0} degrees\n'.format(meanClippedRAOffset))
    outLines.append('Std. dev. 3-sigma-clipped RA offset (1 - 2): {0} degrees\n\n'.format(stdClippedRAOffset))

    outLines.append('Mean Dec offset (1 - 2): {0} degrees\n'.format(meanDecOffset))
    outLines.append('Std. dev. Dec offset (1 - 2): {0} degrees\n'.format(stdDecOffset))
    outLines.append('Mean 3-sigma-clipped Dec offset (1 - 2): {0} degrees\n'.format(meanClippedDecOffset))
    outLines.append('Std. dev. 3-sigma-clipped Dec offset (1 - 2): {0} degrees\n\n'.format(stdClippedDecOffset))

    fileName = outDir+'stats.txt'
    statsFile = open(fileName, 'w')
    statsFile.writelines(outLines)
    statsFile.close()

    return stats



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation implements ungrouping of patches
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging

log = logging.getLogger('LSMTool.UNGROUP')
log.debug('Loading UNGROUP module.')


def run(step, parset, LSM):

    outFile = parset.getString('.'.join(["LSMTool.Steps", step, "OutFile"]), '')

    try:
        LSM.ungroup()
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    # Write to outFile
    if outFile != '' and result == 0:
        LSM.write(outFile, clobber=True)

    return result



# -*- coding: utf-8 -*-
#
# This module initializes the LSMTool module
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""The load() convenience function is used to load a sky model file into a
SkyModel object.

.. moduleauthor:: David Rafferty <drafferty@hs.uni-hamburg.de>

"""

from ._version import changelog
from . import _logging as logger
logger.setLevel('info')


def load(fileName, beamMS=None, VOPosition=None, VORadius=None):
    """
    Loads a sky model from a file or VO service and returns a SkyModel object.

    Parameters
    ----------
    fileName : str
        Input ASCII file from which the sky model is read (must respect the
        makesourcedb format), name of VO service to query (must be one of
        'WENSS', 'NVSS', 'TGSS', or 'GSM'), or dict (single source only)
    beamMS : str, optional
        Measurement set from which the primary beam will be estimated. A
        column of attenuated Stokes I fluxes will be added to the table
    VOPosition : list of floats
        A list specifying a new position as [RA, Dec] in either makesourcedb
        format (e.g., ['12:23:43.21', '+22.34.21.2']) or in degrees (e.g.,
        [123.2312, 23.3422]) for a cone search
    VORadius : float or str, optional
        Radius in degrees (if float) or 'value unit' (if str; e.g.,
        '30 arcsec') for cone search region in degrees

    Returns
    -------
    SkyModel object
        A SkyModel object that stores the sky model and provides methods for
        accessing it.

    Examples
    --------
    Load a sky model from a makesourcedb-formated file::

        >>> import lsmtool
        >>> s = lsmtool.load('sky.model')

    Load a sky model with a beam MS so that apparent fluxes will
    be available (in addition to intrinsic fluxes)::

        >>> s = lsmtool.load('sky.model', 'SB100.MS')

    Load a sky model from the WENSS using all sources within 5 degrees of the
    position RA = 212.8352792, Dec = 52.202644::

        >>> s = lsmtool.load('WENSS', VOPosition=[212.8352792, 52.202644],
            VOradius=5.0)

    Load a sky model from a dictionary defining the source::

        >>> source = {'Name':'src1', 'Type':'POINT', 'Ra':'12:32:10.1',
            'Dec':'23.43.21.21', 'I':2.134}
        >>> s = lsmtool.load(source)

    """
    from .skymodel import SkyModel

    return SkyModel(fileName, beamMS=beamMS, VOPosition=VOPosition,
        VORadius=VORadius)



# -*- coding: utf-8 -*-
#
# Defines meanshift grouper functions used by the group operation. Based on
# implementation of Francesco de Gasperin (see
# https://github.com/revoltek/LiLF/blob/master/lib_dd.py)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
import numpy as np
import logging
log = logging.getLogger('LSMTool.Group')


class Grouper(object):
    """
    Class to group a list of coordinates and fluxes by the meanshift algorighm.
    Based on: http://www.chioka.in/meanshift-algorithm-for-the-rest-of-us-python/

    Parameters
    ----------
    coords : list
        List of (x, y) pixel coordinates for source positions
    fluxes: list
        total flux density for each source
    kernel_size : float
        Kernel size in pixels
    n_iterations : int
        Number of iterations
    look_distance : float
        Look distance in pixels
    grouping_distance : float
        Grouping distance in pixels
    """
    def __init__(self, coords, fluxes, kernel_size, n_iterations, look_distance,
                 grouping_distance):
        self.coords = np.array(coords)
        self.fluxes = fluxes
        self.kernel_size = kernel_size
        self.n_iterations = n_iterations
        self.look_distance = look_distance
        self.grouping_distance = grouping_distance
        self.past_coords = [np.copy(self.coords)]
        self.clusters = []

    def euclid_distance(self, coord, coords):
        """
        Simple distance from coord to all coords
        """
        return np.sqrt(np.sum((coord - coords)**2, axis=1))

    def neighbourhood_points(self, centroid, coords, max_distance):
        """
        Find close points, this reduces the load
        """
        distances = self.euclid_distance(centroid, coords)
        return np.where(distances < max_distance)

    def gaussian_kernel(self, distance):
        """
        Defines a Gaussian kernel
        """
        return (1/(self.kernel_size*np.sqrt(2*np.pi))) * np.exp(-0.5*((distance / self.kernel_size))**2)

    def run(self):
        """
        Run the algorithm
        """
        for it in range(self.n_iterations):
            log.info("Starting iteration %i" % it)

            for i, x in enumerate(self.coords):
                # Step 1. For each datapoint x in X, find the neighbouring points N(x) of x.
                idx_neighbours = self.neighbourhood_points(x, self.coords, max_distance=self.look_distance)

                # Step 2. For each datapoint x in X, calculate the mean shift m(x).
                distances = self.euclid_distance(self.coords[idx_neighbours], x)
                weights = self.gaussian_kernel(distances)
                weights *= self.fluxes[idx_neighbours]
                numerator = np.sum(weights[:, np.newaxis] * self.coords[idx_neighbours], axis=0)
                denominator = np.sum(weights)
                new_x = numerator / denominator

                # Step 3. For each datapoint x in X, update x <- m(x).
                self.coords[i] = new_x

            self.past_coords.append(np.copy(self.coords))

            # if things change little, break
            if it > 1 and np.max(self.euclid_distance(self.coords, self.past_coords[-2])) < self.grouping_distance/2.0:
                break

    def grouping(self):
        """
        Take the last coords set and group sources nearby, then return a list of lists.
        Each list has the index of one cluster.
        """
        coords_to_check = np.copy(self.coords)
        while len(coords_to_check) > 0:
            idx_cluster = self.neighbourhood_points(coords_to_check[0], self.coords, max_distance=self.grouping_distance)
            idx_cluster_to_remove = self.neighbourhood_points(coords_to_check[0], coords_to_check, max_distance=self.grouping_distance)

            # remove all coords of this clusters from the global list
            mask = np.ones(coords_to_check.shape[0], dtype=bool)
            mask[idx_cluster_to_remove] = False
            coords_to_check = coords_to_check[mask]

            # save this cluster indexes
            self.clusters.append(idx_cluster)

        log.info('Creating %i groups.' % len(self.clusters))
        return self.clusters



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This operation implements joining of two sky models
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging

log = logging.getLogger('LSMTool.CONCATENATE')
log.debug('Loading CONCATENATE module.')


def run(step, parset, LSM):

    outFile = parset.getString('.'.join(["LSMTool.Steps", step, "OutFile"]), '' )
    skyModel2 = parset.getString('.'.join(["LSMTool.Steps", step, "Skymodel2"]), '' )
    matchBy = parset.getString('.'.join(["LSMTool.Steps", step, "MatchBy"]), 'name' )
    radius = parset.getString('.'.join(["LSMTool.Steps", step, "Radius"]), '0.1' )
    keep = parset.getString('.'.join(["LSMTool.Steps", step, "Keep"]), 'all' )
    inheritPatches = parset.getBool('.'.join(["LSMTool.Steps", step, "InheritPatches"]), False )

    try:
        concatenate(LSM, skyModel2, matchBy, radius, keep, inheritPatches)
        result = 0
    except Exception as e:
        log.error(e)
        result = 1

    # Write to outFile
    if outFile != '' and result == 0:
        LSM.write(outFile, clobber=True)

    return result


def concatenate(LSM1, LSM2, matchBy='name', radius=0.1, keep='all',
    inheritPatches=False):
    """
    Concatenate two sky models

    Note that the mergesourcedb tool performs a similar function (but on
    SourceDB data sets, not on sky model files; see
    http://www.lofar.org/operations/doku.php?id=engineering:software:tools:makesourcedb#format_string
    for details).

    Parameters
    ----------
    LSM1 : SkyModel object
        Parent sky model
    LSM2 : SkyModel object
        Sky model to concatenate with the parent sky model
    matchBy : str, optional
        Determines how duplicate sources are determined:
        - 'name' => duplicates are identified by name
        - 'position' => duplicates are identified by radius. Sources within the
            radius specified by the radius parameter are considered duplicates
    radius : float or str, optional
        Radius in degrees (if float) or 'value unit' (if str; e.g., '30 arcsec')
        for matching when matchBy='position'
    keep : str, optional
        Determines how duplicates are treated:
        - 'all' => all duplicates are kept; those with identical names are re-
            named
        - 'from1' => duplicates kept are those from sky model 1 (LSM1)
        - 'from2' => duplicates kept are those from sky model 2 (LSM2)
    inheritPatches : bool, optional
        If True, duplicates inherit the patch name from the parent sky model. If
        False, duplicates keep their own patch names.

    Examples
    --------
    Concatenate two sky models, identifying duplicates by matching to the source
    names. When duplicates are found, keep the source from the parent sky model
    and discard the duplicate from second sky model (this might be useful when
    merging two sky models that have some overlap)::

        >>> LSM1 = lsmtool.load('sky1.model')
        >>> LSM2 = lsmtool.load('sky2.model')
        >>> concatenate(LSM1, LSM2, matchBy='name', keep='from1')

    Concatenate two sky models, identifying duplicates by matching to the source
    positions within a radius of 10 arcsec. When duplicates are found, keep the
    source from the second sky model and discard the duplicate from the parent
    sky model (this might be useful when replacing parts of a low-resolution
    sky model with a high-resolution one)::

        >>> LSM2 = lsmtool.load('high_res_sky.model')
        >>> concatenate(LSM1, LSM2, matchBy='position', radius=10.0/3600.0,
            keep='from2')

    """
    from astropy.table import vstack, Column
    from ..operations_lib import matchSky
    from ..skymodel import SkyModel
    import numpy as np

    if type(LSM2) is str:
        LSM2 = SkyModel(LSM2)

    if len(LSM1) == 0:
        log.info('Parent sky model is empty. Concatenated sky model is '
            'copy of secondary sky model.')
        LSM1.table = LSM2.table
        LSM1._updateGroups()
        LSM1._info()
        return
    if len(LSM2) == 0:
        log.info('Secondary sky model is empty. Parent sky model left '
            'unaltered.')
        return

    if (LSM1.hasPatches and not LSM2.hasPatches):
         LSM2.group('every')
    if (LSM2.hasPatches and not LSM1.hasPatches):
         LSM2.ungroup()

    # Make sure spectral index entries are of same length
    if 'SpectralIndex' in LSM1.getColNames() and 'SpectralIndex' in LSM2.getColNames():
        len1 = len(LSM1.getColValues('SpectralIndex')[0])
        len2 = len(LSM2.getColValues('SpectralIndex')[0])

        if len1 < len2:
            oldspec = LSM1.getColValues('SpectralIndex')
            newspec = []
            for specind in oldspec:
                speclist = specind.tolist()
                while len(speclist) < len2:
                    speclist.append(0.0)
                newspec.append(speclist)
            specCol = Column(name='SpectralIndex', data=np.array(newspec, dtype=float))
            specIndx = LSM1.table.keys().index('SpectralIndex')
            LSM1.table.remove_column('SpectralIndex')
            LSM1.table.add_column(specCol, index=specIndx)

        elif len1 > len2:
            oldspec = LSM2.getColValues('SpectralIndex')
            newspec = []
            for specind in oldspec:
                speclist = specind.tolist()
                while len(speclist) < len1:
                    speclist.append(0.0)
                newspec.append(speclist)
            specCol = Column(name='SpectralIndex', data=np.array(newspec, dtype=float))
            specIndx = LSM2.table.keys().index('SpectralIndex')
            LSM2.table.remove_column('SpectralIndex')
            LSM2.table.add_column(specCol, index=specIndx)

    # Fill masked values and merge defaults and RA, Dec formaters
    table1 = LSM1.table.filled()
    table2 = LSM2.table.filled()
    for entry in table1.meta:
        if LSM1._verifyColName(entry, quiet=True) is not None:
            if entry in table2.meta.keys():
                table1.meta[entry] = table2.meta[entry]
    table1['Ra'].format = table2['Ra'].format
    table1['Dec'].format = table2['Dec'].format

    # Now concatenate the tables
    if matchBy.lower() == 'name':
        LSM1.table = vstack([table1, table2], metadata_conflicts='silent')
    elif matchBy.lower() == 'position':
        matches1, matches2 = matchSky(LSM1, LSM2, radius=radius)
        matchCol1 = np.array(range(len(LSM1)))
        matchCol2 = np.array(range(len(LSM2))) + len(LSM1)

        # Set values to be the same for the matches
        matchCol2[matches2] = matchCol1[matches1]

        # Now add columns and stack
        col1 = Column(name='match', data=matchCol1)
        col2 = Column(name='match', data=matchCol2)
        table1.add_column(col1)
        table2.add_column(col2)
        LSM1.table = vstack([table1, table2], metadata_conflicts='silent')

    if keep == 'from1' or keep == 'from2':
        # Remove any duplicates
        if matchBy.lower() == 'name':
            colName = 'Name'
        elif matchBy.lower() == 'position':
            colName = 'match'
        else:
            raise ValueError('Invalid matchBy parameter.')
        vals = LSM1.table[colName]
        for val in vals:
            valsCur = LSM1.table[colName]
            toRemove = []
            indx = np.where(valsCur == val)[0]
            if len(indx) > 1:
                if keep == 'from1':
                    toRemove.append(indx[1:])
                else:
                    toRemove.append(indx[0])
                    if inheritPatches and LSM1.hasPatches:
                        LSM1.table['Patch'][indx[1:]] = LSM1.table['Patch'][indx[0]]
                LSM1.table.remove_rows(toRemove)

    # Rename any duplicates
    check_duplicates = True
    while check_duplicates:
        names = LSM1.getColValues('Name')
        check_duplicates = False
        for name in set(names):
            indx = np.where(names == name)[0]
            if len(indx) > 1:
                check_duplicates = True
                LSM1.table['Name'][indx[0]] = name + '_1'
                LSM1.table['Name'][indx[1]] = name + '_2'

    if matchBy.lower() == 'position':
        LSM1.table.remove_column('match')

    LSM1._updateGroups()
    history = "matchBy = '{0}', ".format(matchBy)
    if matchBy.lower() == 'position':
        history += "radius = {0}".format(radius)
    history += " keep = '{0}'".format(keep)
    LSM1._addHistory("CONCATENATE ({0})".format(history))
    LSM1._info()

